# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the definitions and registrations for the various packets used in the server
"""

# Notes:
# Here Packets are registered with the discouraged register_packet() notation.
# This is due to these packets all being cdef. This means you can not assign to
# them, and hence not use decorators on them.
#
# Other things that should probably be done here is using cython.freelist(n) to
# speed up allocation for packets

from pyspades.common import encode, decode
from pyspades.constants import NEUTRAL_TEAM, CTF_MODE, TC_MODE
from pyspades.loaders cimport Loader
from pyspades.bytes cimport ByteReader, ByteWriter
from pyspades.packet import register_packet

cimport cython

cdef inline float limit(float a):
    if a > 512.0:
        return 512.0
    elif a < 0.0:
        return 0.0
    return a

cdef inline void read_position(ByteReader reader, float * x, float * y,
                               float * z):
    x[0] = reader.readFloat(False)
    y[0] = reader.readFloat(False)
    z[0] = reader.readFloat(False)

cdef inline void write_position(ByteWriter writer, float x, float y, float z):
    writer.writeFloat(x, False)
    writer.writeFloat(y, False)
    writer.writeFloat(z, False)

cdef inline unsigned int read_color(ByteReader reader):
    cdef unsigned char r, g, b
    b = reader.readByte(True)
    g = reader.readByte(True)
    r = reader.readByte(True)
    return (b | (g << 8) | (r << 16))

cdef inline void write_color(ByteWriter writer, unsigned int value):
    writer.writeByte(value & 0xFF)
    writer.writeByte((value >> 8) & 0xFF)
    writer.writeByte((value >> 16) & 0xFF)

import itertools
id_iter = itertools.count()

cdef class PositionData(Loader):
    id = 0

    cdef public:
        float x, y, z

    def set(self, pos):
        cdef float x, y, z
        x, y, z = pos
        self.x = x
        self.y = y
        self.z = z

    cpdef read(self, ByteReader reader):
        read_position(reader, &self.x, &self.y, &self.z)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        write_position(writer, self.x, self.y, self.z)

register_packet(PositionData)

cdef class OrientationData(Loader):
    id = 1

    cdef public:
        float x, y, z

    def set(self, pos):
        cdef float x, y, z
        x, y, z = pos
        self.x = x
        self.y = y
        self.z = z

    cpdef read(self, ByteReader reader):
        self.x = reader.readFloat(False)
        self.y = reader.readFloat(False)
        self.z = reader.readFloat(False)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeFloat(self.x, False)
        writer.writeFloat(self.y, False)
        writer.writeFloat(self.z, False)

register_packet(OrientationData)

cdef class WorldUpdate(Loader):
    id = 2

    cdef public:
        list items

    cpdef read(self, ByteReader reader):
        cdef list items = []
        self.items = items
        for _ in range(32):
            p_x = reader.readFloat(False)
            p_y = reader.readFloat(False)
            p_z = reader.readFloat(False)
            o_x = reader.readFloat(False)
            o_y = reader.readFloat(False)
            o_z = reader.readFloat(False)
            items.append(((p_x, p_y, p_z), (o_x, o_y, o_z)))

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        cdef tuple item
        for item in self.items:
            (p_x, p_y, p_z), (o_x, o_y, o_z) = item
            writer.writeFloat(p_x, False)
            writer.writeFloat(p_y, False)
            writer.writeFloat(p_z, False)
            writer.writeFloat(o_x, False)
            writer.writeFloat(o_y, False)
            writer.writeFloat(o_z, False)

register_packet(WorldUpdate)

cdef class InputData(Loader):
    id = 3
    cdef public:
        int player_id
        bint up, down, left, right, jump, crouch, sneak, sprint

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        cdef int firstByte = reader.readByte(True)
        self.up = (firstByte >> 0) & 1
        self.down = (firstByte >> 1) & 1
        self.left = (firstByte >> 2) & 1
        self.right = (firstByte >> 3) & 1
        self.jump = (firstByte >> 4) & 1
        self.crouch = (firstByte >> 5) & 1
        self.sneak = (firstByte >> 6) & 1
        self.sprint = (firstByte >> 7) & 1

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        cdef int byte
        byte = (self.up | (self.down << 1) | (self.left << 2) |
            (self.right << 3) | (self.jump << 4) | (self.crouch << 5) |
            (self.sneak << 6) | (self.sprint << 7))
        writer.writeByte(byte, True)

register_packet(InputData)

cdef class WeaponInput(Loader):
    id = 4

    cdef public:
        bint primary, secondary
        int player_id

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        cdef unsigned char byte = reader.readByte(True)
        self.primary = byte & 1
        self.secondary = (byte >> 1) & 1

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        cdef unsigned char byte = self.primary | (self.secondary << 1)
        writer.writeByte(byte, True)


register_packet(WeaponInput)

cdef class HitPacket(Loader):
    id = 5

    cdef public:
        int player_id, value

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.value = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.value, True)

register_packet(HitPacket, server=False)

cdef class SetHP(Loader):
    id = 5

    cdef public:
        int hp, not_fall
        float source_x, source_y, source_z

    cpdef read(self, ByteReader reader):
        self.hp = reader.readByte(True)
        # FALL = 0, WEAPON = 1
        self.not_fall = reader.readByte(True)
        self.source_x = reader.readFloat(False)
        self.source_y = reader.readFloat(False)
        self.source_z = reader.readFloat(False)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.hp, True)
        writer.writeByte(self.not_fall, True)
        writer.writeFloat(self.source_x, False)
        writer.writeFloat(self.source_y, False)
        writer.writeFloat(self.source_z, False)

register_packet(SetHP, client=False)

cdef class GrenadePacket(Loader):
    id = 6

    cdef public:
        int player_id
        float value
        tuple position, velocity

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.value = reader.readFloat(False)
        self.position = (reader.readFloat(False), reader.readFloat(False),
            reader.readFloat(False))
        self.velocity = (reader.readFloat(False), reader.readFloat(False),
            reader.readFloat(False))

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeFloat(self.value, False)
        for value in self.position:
            writer.writeFloat(value, False)
        for value in self.velocity:
            writer.writeFloat(value, False)

register_packet(GrenadePacket)

@cython.freelist(8)
cdef class SetTool(Loader):
    id = 7

    cdef public:
        int player_id, value

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.value = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.value, True)

register_packet(SetTool)

cdef class SetColor(Loader):
    id = 8

    cdef public:
        unsigned int value, player_id

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.value = read_color(reader)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        write_color(writer, self.value)

register_packet(SetColor)

cdef class ExistingPlayer(Loader):
    id = 9

    cdef public:
        int player_id, team, weapon, tool, kills
        unsigned int color
        object name

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.team = reader.readByte(False)
        self.weapon = reader.readByte(True)
        self.tool = reader.readByte(True)
        self.kills = reader.readInt(True, False)
        self.color = read_color(reader)
        self.name = decode(reader.readString()) # 16 bytes

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.team, False)
        writer.writeByte(self.weapon, True)
        writer.writeByte(self.tool, True)
        writer.writeInt(self.kills, True, False)
        write_color(writer, self.color)
        writer.writeString(encode(self.name))

register_packet(ExistingPlayer)

cdef class ShortPlayerData(Loader):
    id = 10

    cdef public:
        int player_id, team, weapon

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.team = reader.readByte(False)
        self.weapon = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.team, False)
        writer.writeByte(self.weapon, True)

register_packet(ShortPlayerData)

cdef class MoveObject(Loader):
    id = 11

    cdef public:
        unsigned int object_type, state
        float x, y, z

    cpdef read(self, ByteReader reader):
        self.object_type = reader.readByte(True)
        self.state = reader.readByte(True)
        self.x = reader.readFloat(False)
        self.y = reader.readFloat(False)
        self.z = reader.readFloat(False)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.object_type, True)
        writer.writeByte(self.state, True)
        writer.writeFloat(self.x, False)
        writer.writeFloat(self.y, False)
        writer.writeFloat(self.z, False)

register_packet(MoveObject)

@cython.freelist(8)
cdef class CreatePlayer(Loader):
    id = 12

    cdef public:
        unsigned int player_id, weapon
        int team
        float x, y, z
        object name

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.weapon = reader.readByte(True)
        self.team = reader.readByte(False)
        read_position(reader, &self.x, &self.y, &self.z)
        self.name = decode(reader.readString())

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.weapon, True)
        writer.writeByte(self.team, False)
        write_position(writer, self.x, self.y, self.z)
        writer.writeString(encode(self.name))

register_packet(CreatePlayer)

cdef class BlockAction(Loader):
    id = 13

    cdef public:
        int x, y, z, value, player_id

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.value = reader.readByte(True)
        self.x = reader.readInt(False, False)
        self.y = reader.readInt(False, False)
        self.z = reader.readInt(False, False)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.value, True)
        writer.writeInt(self.x, False, False)
        writer.writeInt(self.y, False, False)
        writer.writeInt(self.z, False, False)

register_packet(BlockAction)

cdef class BlockLine(Loader):
    id = 14

    cdef public:
        int player_id
        int x1, y1, z1
        int x2, y2, z2

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.x1 = reader.readInt(False, False)
        self.y1 = reader.readInt(False, False)
        self.z1 = reader.readInt(False, False)
        self.x2 = reader.readInt(False, False)
        self.y2 = reader.readInt(False, False)
        self.z2 = reader.readInt(False, False)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeInt(self.x1, False, False)
        writer.writeInt(self.y1, False, False)
        writer.writeInt(self.z1, False, False)
        writer.writeInt(self.x2, False, False)
        writer.writeInt(self.y2, False, False)
        writer.writeInt(self.z2, False, False)

register_packet(BlockLine)

cdef class CTFState(Loader):
    id = CTF_MODE # this is not a real a packet, it sent as part of the StateData packet
                  # data

    cdef public:
        unsigned int team1_score, team2_score, cap_limit
        bint team1_has_intel, team2_has_intel
        unsigned int team1_carrier, team2_carrier
        float team1_flag_x, team1_flag_y, team1_flag_z
        float team2_flag_x, team2_flag_y, team2_flag_z
        float team1_base_x, team1_base_y, team1_base_z
        float team2_base_x, team2_base_y, team2_base_z

    cpdef read(self, ByteReader reader):
        self.team1_score = reader.readByte(True)
        self.team2_score = reader.readByte(True)
        self.cap_limit = reader.readByte(True)

        cdef int intel_flags = reader.readByte(True)
        self.team1_has_intel = intel_flags & 1
        self.team2_has_intel = (intel_flags >> 1) & 1

        if self.team2_has_intel:
            self.team1_carrier = reader.readByte(True)
            reader.skipBytes(11)
        else:
            read_position(reader, &self.team1_flag_x, &self.team1_flag_y,
                &self.team1_flag_z)

        if self.team1_has_intel:
            self.team2_carrier = reader.readByte(True)
            reader.skipBytes(11)
        else:
            read_position(reader, &self.team2_flag_x, &self.team2_flag_y,
                &self.team2_flag_z)

        read_position(reader, &self.team1_base_x, &self.team1_base_y,
            &self.team1_base_z)

        read_position(reader, &self.team2_base_x, &self.team2_base_y,
            &self.team2_base_z)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.team1_score, True)
        writer.writeByte(self.team2_score, True)
        writer.writeByte(self.cap_limit, True)
        cdef int intel_flags = (self.team1_has_intel | (
            self.team2_has_intel << 1))

        writer.writeByte(intel_flags, True)

        if self.team2_has_intel:
            writer.writeByte(self.team1_carrier, True)
            writer.pad(11)
        else:
            write_position(writer, self.team1_flag_x, self.team1_flag_y,
                self.team1_flag_z)

        if self.team1_has_intel:
            writer.writeByte(self.team2_carrier, True)
            writer.pad(11)
        else:
            write_position(writer, self.team2_flag_x, self.team2_flag_y,
                self.team2_flag_z)

        write_position(writer, self.team1_base_x, self.team1_base_y,
            self.team1_base_z)

        write_position(writer, self.team2_base_x, self.team2_base_y,
            self.team2_base_z)

DEF MAX_TERRITORIES = 16
DEF TERRITORY_SIZE = 4*3+1
DEF TERRITORY_DATA = MAX_TERRITORIES * TERRITORY_SIZE

cdef class Territory(Loader):
    cdef public:
        float x, y, z
        unsigned int state

    cpdef read(self, ByteReader reader):
        read_position(reader, &self.x, &self.y, &self.z)
        self.state = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        write_position(writer, self.x, self.y, self.z)
        writer.writeByte(self.state, True)

cdef class ObjectTerritory(Loader):
    cdef public:
        object item

    cpdef write(self, ByteWriter writer):
        write_position(writer, self.item.x, self.item.y, self.item.z)
        team = self.item.team
        cdef int state
        if team is None:
            state = NEUTRAL_TEAM
        else:
            state = team.id
        writer.writeByte(state, True)

cdef class TCState(Loader):
    id = TC_MODE # this is not a real a packet, it sent as part of the StateData packet
                 # data

    cdef public:
        list territories

    cpdef read(self, ByteReader reader):
        self.territories = []
        cdef unsigned int count = reader.readByte(True)
        for _ in range(count):
            self.territories.append(Territory(reader))

    def set_entities(self, items):
        self.territories = []
        for item in items:
            territory = ObjectTerritory()
            territory.item = item
            self.territories.append(territory)

    cpdef write(self, ByteWriter writer):
        cdef Loader territory
        writer.writeByte(len(self.territories), True)
        for territory in self.territories:
            territory.write(writer)
        writer.pad((MAX_TERRITORIES - len(self.territories)) * TERRITORY_SIZE)

modes = {
    CTF_MODE : CTFState,
    TC_MODE : TCState
}

cdef inline tuple read_team_color(ByteReader reader):
    b = reader.readByte(True)
    g = reader.readByte(True)
    r = reader.readByte(True)
    return (r, g, b)

cdef inline void write_team_color(ByteWriter writer, tuple color):
    r, g, b = color
    writer.writeByte(b, True)
    writer.writeByte(g, True)
    writer.writeByte(r, True)

cdef class StateData(Loader):
    id = 15

    cdef public:
        int player_id
        tuple fog_color
        tuple team1_color
        tuple team2_color
        Loader state
        object team1_name, team2_name

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.fog_color = read_team_color(reader)
        self.team1_color = read_team_color(reader)
        self.team2_color = read_team_color(reader)
        self.team1_name = decode(reader.readString(10))
        self.team2_name = decode(reader.readString(10))
        cdef int mode = reader.readByte(True)
        self.state = modes[mode](reader)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        write_team_color(writer, self.fog_color)
        write_team_color(writer, self.team1_color)
        write_team_color(writer, self.team2_color)
        writer.writeString(encode(self.team1_name), 10)
        writer.writeString(encode(self.team2_name), 10)
        writer.writeByte(self.state.id, True)
        self.state.write(writer)

register_packet(StateData)

cdef class KillAction(Loader):
    id = 16

    cdef public:
        int player_id, killer_id, kill_type, respawn_time

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.killer_id = reader.readByte(True)
        self.kill_type = reader.readByte(True)
        self.respawn_time = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.killer_id, True)
        writer.writeByte(self.kill_type, True)
        writer.writeByte(self.respawn_time, True)

register_packet(KillAction)

cdef class ChatMessage(Loader):
    id = 17

    cdef public:
        unsigned int player_id, chat_type
        object value

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.chat_type = reader.readByte(True)
        self.value = decode(reader.readString())

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.chat_type, True)
        writer.writeString(encode(self.value))

register_packet(ChatMessage)

cdef class MapStart(Loader):
    id = 18

    cdef public:
        unsigned int size

    cpdef read(self, ByteReader reader):
        self.size = reader.readInt(True, False)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeInt(self.size, True, False)

register_packet(MapStart)

cdef class MapChunk(Loader):
    id = 19

    cdef public:
        object data

    cpdef read(self, ByteReader reader):
        self.data = reader.read()

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.write(self.data)

register_packet(MapChunk)

@cython.freelist(8)
cdef class PlayerLeft(Loader):
    id = 20

    cdef public:
        int player_id

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)

register_packet(PlayerLeft)

cdef class TerritoryCapture(Loader):
    id = 21

    cdef public:
        unsigned int object_index, winning, state

    cpdef read(self, ByteReader reader):
        self.object_index = reader.readByte(True)
        self.winning = reader.readByte(True)
        self.state = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.object_index, True)
        writer.writeByte(self.winning, True)
        writer.writeByte(self.state, True)

register_packet(TerritoryCapture)

cdef class ProgressBar(Loader):
    id = 22

    cdef public:
        unsigned int object_index, capturing_team
        int rate
        float progress

    cpdef read(self, ByteReader reader):
        self.object_index = reader.readByte(True)
        self.capturing_team = reader.readByte(True)
        self.rate = reader.readByte(False)
        self.progress = reader.readFloat(False)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.object_index, True)
        writer.writeByte(self.capturing_team, True)
        writer.writeByte(self.rate, False)
        writer.writeFloat(self.progress, False)

register_packet(ProgressBar)

@cython.freelist(8)
cdef class IntelCapture(Loader):
    id = 23

    cdef public:
        int player_id
        bint winning

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.winning = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.winning, True)

register_packet(IntelCapture)

cdef class IntelPickup(Loader):
    id = 24

    cdef public:
        int player_id

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)

register_packet(IntelPickup)

cdef class IntelDrop(Loader):
    id = 25

    cdef public:
        int player_id
        float x, y, z

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        read_position(reader, &self.x, &self.y, &self.z)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        write_position(writer, self.x, self.y, self.z)

register_packet(IntelDrop)

cdef class Restock(Loader):
    id = 26

    cdef public:
        int player_id

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)

register_packet(Restock)

@cython.freelist(8)
cdef class FogColor(Loader):
    id = 27

    cdef public:
        int color

    cpdef read(self, ByteReader reader):
        self.color = reader.readInt(True, False) >> 8

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeInt(self.color << 8, True, False)

register_packet(FogColor)

@cython.freelist(8)
cdef class WeaponReload(Loader):
    id = 28

    cdef public:
        int player_id, clip_ammo, reserve_ammo

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.clip_ammo = reader.readByte(True)
        self.reserve_ammo = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.clip_ammo, True)
        writer.writeByte(self.reserve_ammo, True)

register_packet(WeaponReload)

cdef class ChangeTeam(Loader):
    id = 29

    cdef public:
        int player_id, team

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.team = reader.readByte(False)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.team, False)

register_packet(ChangeTeam)

cdef class ChangeWeapon(Loader):
    id = 30

    cdef public:
        int player_id, weapon

    cpdef read(self, ByteReader reader):
        self.player_id = reader.readByte(True)
        self.weapon = reader.readByte(True)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeByte(self.player_id, True)
        writer.writeByte(self.weapon, True)

register_packet(ChangeWeapon)

cdef class HandShakeInit(Loader):
    id = 31

    cpdef read(self, ByteReader reader):
        pass

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)
        writer.writeInt(42, True)

register_packet(HandShakeInit)

cdef class HandShakeReturn(Loader):
    id = 32

    cdef public:
        int success

    cpdef read(self, ByteReader reader):
        self.success = int(reader.readInt(True) == 42)

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)

register_packet(HandShakeReturn)

cdef class VersionRequest(Loader):
    id = 33

    cpdef read(self, ByteReader reader):
        pass

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)

register_packet(VersionRequest)

cdef class VersionResponse(Loader):
    id = 34

    cdef public:
        str client
        tuple version
        str os_info

    cpdef read(self, ByteReader reader):
        magic_no = reader.readByte(True)
        self.client = chr(magic_no)
        self.version = (
            reader.readByte(True),
            reader.readByte(True),
            reader.readByte(True),
        )
        self.os_info = decode(reader.readString())

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)

register_packet(VersionResponse)


cdef class ProtocolExtensionInfo(Loader):
    """packet used to exchange the list of supported protocol extensions between
    server and client

    extensions is a list of (extension_id, version) tuples
    """
    id = 60

    cdef public:
        list extensions

    cpdef read(self, ByteReader reader):
        extension_count = reader.readByte(True)
        extensions = []
        for _ in range(extension_count):
            extensions.append(
                (reader.readByte(True), reader.readByte(True))
            )
        self.extensions = extensions

    cpdef write(self, ByteWriter writer):
        writer.writeByte(self.id, True)

        writer.writeByte(len(self.extensions), True)

        for ext in self.extensions:
            writer.writeByte(ext[0])
            writer.writeByte(ext[1])

register_packet(ProtocolExtensionInfo)



"""
Restocks the user when reloading / throwing a nade.

Commands
^^^^^^^^

* ``/toggledemo`` toggles demolition

.. codeauthor:: learn_more (MIT LICENSE)
"""

from piqueserver.commands import command, admin

DEMOLITION_ENABLED_AT_ROUND_START = False


@command(admin_only=True)
def toggledemo(connection):
    connection.protocol.demolitionEnabled = not connection.protocol.demolitionEnabled
    message = 'Demolition is now disabled'
    if connection.protocol.demolitionEnabled:
        message = 'Demolition is now enabled'
    connection.protocol.send_chat(message, irc=True)
    return 'ok :)'


def apply_script(protocol, connection, config):
    class DemolitionProtocol(protocol):
        demolitionEnabled = DEMOLITION_ENABLED_AT_ROUND_START

        def on_map_change(self, map):
            self.demolitionEnabled = DEMOLITION_ENABLED_AT_ROUND_START
            return protocol.on_map_change(self, map)

    class DemolitionConnection(connection):

        def _on_reload(self):
            if self.protocol.demolitionEnabled:
                self.refill()
            return connection._on_reload(self)

        def on_grenade_thrown(self, grenade):
            if self.protocol.demolitionEnabled:
                self.refill()
            return connection.on_grenade_thrown(self, grenade)

        def on_spawn(self, pos):
            if self.protocol.demolitionEnabled:
                self.send_chat(
                    'You are the demolition man, grenades & ammo will be restocked!')
            return connection.on_spawn(self, pos)

    return DemolitionProtocol, DemolitionConnection



'''
Gets a player's location info using a geoip database.
Location is formatted as a list of regions ordered by hierarchy (i.e. City, Region, Country)

.. note::
  This script depends on `geoip2` package and `piqueserver --update-geoip` needs to be executed after installing the package.

Commands
^^^^^^^^
* ``/from ``         get active player's location
* ``/from <player>`` get player's location info

.. codeauthor:: ?
'''

import os
from piqueserver.commands import (command, restrict, get_player, target_player)
from piqueserver.config import config

# optional commands
try:
    import geoip2.database
    import geoip2.errors
    database = geoip2.database.Reader(os.path.join(
        config.config_dir, 'data/GeoLite2-City.mmdb'))
except ImportError:
    print("('/from' command disabled. Please install geoip2 to enable.)")
except (IOError, OSError):
    print(
        "('/from' command disabled due to missing GeoIP database. Run 'piqueserver --update-geoip' to install.)"
    )
finally:

    @command('from', admin_only=True)
    @target_player
    def where_from(connection, player):
        # Query database
        try:
            record = database.city(player.address[0])
        except geoip2.errors.GeoIP2Error:
            return 'Player location could not be determined.'

        # Extract info
        raw_items = (record.city, *reversed(record.subdivisions),
                     record.country)

        items = (raw_item.name for raw_item in raw_items if raw_item.name is not None)

        return '%s is from %s (%s)' % (player.name, ', '.join(items),
                                       record.country.iso_code)


def apply_script(protocol, connection, config):
    return protocol, connection



"""
Kicks a player if inactive for too long.

Options
^^^^^^^

.. code-block:: guess

   [afk]
   time_limit = "15min"

.. codeauthor:: hompy
"""

from operator import attrgetter
from twisted.internet import reactor
from pyspades.common import prettify_timespan
from piqueserver.commands import command, get_player, admin
from piqueserver.config import config, cast_duration

S_AFK_CHECK = '{player} has been inactive for {time}'
S_NO_PLAYERS_INACTIVE = 'No players or connections inactive for {time}'
S_AFK_KICKED = ('{num_players} players kicked, {num_connections} connections '
                'terminated for {time} inactivity')
S_AFK_KICK_REASON = 'Inactive for {time}'

afk_config = config.section('afk')
time_limit_option = afk_config.option('time_limit', default="1hour", cast=cast_duration)

def afk(connection, player):
    player = get_player(connection.protocol, player)
    elapsed = prettify_timespan(reactor.seconds() - player.last_activity, True)
    return S_AFK_CHECK.format(player=player.name, time=elapsed)


@command('kickafk', admin_only=True)
def kick_afk(connection, minutes, amount=None):
    protocol = connection.protocol
    minutes = int(minutes)
    if minutes < 1:
        raise ValueError("Minutes cannot be < 1")
    to_kick = []
    seconds = minutes * 60.0
    minutes_s = prettify_timespan(seconds)
    lower_bound = reactor.seconds() - seconds
    for conn in list(protocol.connections.values()):
        if not conn.admin and conn.last_activity < lower_bound:
            to_kick.append(conn)
    if not to_kick:
        return S_NO_PLAYERS_INACTIVE.format(time=minutes_s)
    to_kick.sort(key=attrgetter('last_activity'))
    to_kick.sort(key=lambda conn: conn.name is None)
    amount = amount or len(to_kick)
    kicks = 0
    for conn in to_kick[:amount]:
        if conn.name:
            conn.afk_kick()
            kicks += 1
        else:
            conn.disconnect()
    message = S_AFK_KICKED.format(
        num_players=kicks,
        num_connections=amount - kicks,
        time=minutes_s)
    protocol.irc_say('* ' + message)
    if connection in protocol.players.values():
        return message


def apply_script(protocol, connection, config):
    time_limit = time_limit_option.get()

    class AFKConnection(connection):
        afk_kick_call = None
        last_activity = None

        def afk_kick(self):
            if self.name:
                time_inactive = reactor.seconds() - self.last_activity
                time_inactive = max(1.0, round(time_inactive / 60.0)) * 60.0
                elapsed = prettify_timespan(time_inactive)
                self.kick(S_AFK_KICK_REASON.format(time=elapsed))
            else:
                self.disconnect()

        def reset_afk_kick_call(self):
            self.last_activity = reactor.seconds()
            if self.afk_kick_call and self.afk_kick_call.active():
                # In Twisted==18.9.0, reset() is broken when using
                # AsyncIOReactor
                # Hence, the following code does not work:
                # self.afk_kick_call.reset(time_limit)
                # It has been temporarily replaced with the equivalent:
                self.afk_kick_call.cancel()
                self.afk_kick_call = reactor.callLater(
                    time_limit, self.afk_kick)

        def on_disconnect(self):
            if self.afk_kick_call and self.afk_kick_call.active():
                self.afk_kick_call.cancel()
            self.afk_kick_call = None
            connection.on_disconnect(self)

        def on_user_login(self, user_type, verbose=True):
            if user_type in ('admin', 'trusted'):
                if self.afk_kick_call and self.afk_kick_call.active():
                    self.afk_kick_call.cancel()
                self.afk_kick_call = None
            return connection.on_user_login(self, user_type, verbose)

        def on_connect(self):
            if not self.local:
                self.afk_kick_call = reactor.callLater(
                    time_limit, self.afk_kick)
            self.last_activity = reactor.seconds()
            return connection.on_connect(self)

        def on_chat(self, value, global_message):
            self.reset_afk_kick_call()
            return connection.on_chat(self, value, global_message)

        def on_orientation_update(self, x, y, z):
            self.reset_afk_kick_call()
            return connection.on_orientation_update(self, x, y, z)

    return protocol, AFKConnection



"""
Allows muting players for a set amount of time.

Commands
^^^^^^^^

* ``/tm <player> <seconds> <reason>`` mute a player for set amount of time *admin only*

.. note::
  Default time 5 minutes, default reason None

.. codeauthor: topologist
"""

from piqueserver.scheduler import Scheduler
from piqueserver.commands import command, get_player, join_arguments


@command('tm', admin_only=True)
def timed_mute(connection, *args):
    protocol = connection.protocol
    if len(args) < 3:
        return '/tm <nick> <time> <reason>'

    nick = args[0]
    time = int(args[1])
    reason = join_arguments(args[2:])
    player = get_player(protocol, nick)

    if time < 0:
        raise ValueError("Time cannot be < 0")

    if not player.mute:
        TimedMute(player, time, reason)
    else:
        return '%s is already muted!' % nick


class TimedMute:
    player = None
    time = None

    def __init__(self, player, time=300, reason='None'):
        if time == 0:
            player.mute = True
            player.protocol.send_chat(
                '%s was muted indefinitely (Reason: %s)' %
                (player.name, reason), irc=True)
            return

        schedule = Scheduler(player.protocol)
        schedule.call_later(time, self.end)
        player.mute_schedule = schedule

        player.protocol.send_chat(
            '%s was muted for %s seconds (Reason: %s)' %
            (player.name, time, reason), irc=True)
        player.mute = True

        self.player = player
        self.time = time

    def end(self):
        self.player.mute = False
        message = '%s was unmuted after %s seconds' % (
            self.player.name, self.time)
        self.player.protocol.send_chat(message, irc=True)


def apply_script(protocol, connection, config):
    class TimedMuteConnection(connection):
        mute_schedule = None

        def on_disconnect(self):
            if self.mute_schedule:
                del self.mute_schedule
            connection.on_disconnect(self)

    return protocol, TimedMuteConnection



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import json
from twisted.internet.task import LoopingCall
from twisted.internet.defer import DeferredList
from twisted.web.client import getPage
from twisted.logger import Logger

from piqueserver.networkdict import NetworkDict
from piqueserver.config import config, cast_duration

log = Logger()

# format is [{"ip" : "1.1.1.1", "reason : "blah"}, ...]


def validate_bansub_config(c):
    if not isinstance(c, list):
        return False

    for item in c:
        if not item.get('url') or not isinstance(item.get('whitelist'), list):
            return False

    return True

bans_config = config.section('bans')
bans_config_urls = bans_config.option('bansubscribe', default=[], validate=validate_bansub_config)
bans_config_interval = bans_config.option('bansubscribe_interval', default="5min",
                                          cast=cast_duration)


class BanManager:
    bans = None
    new_bans = None

    def __init__(self, protocol):
        self.protocol = protocol
        self.urls = [(entry.get('url'), entry.get('whitelist')) for entry in
                     bans_config_urls.get()]
        self.loop = LoopingCall(self.update_bans)
        self.loop.start(bans_config_interval.get(), now=True)

    def update_bans(self):
        self.new_bans = NetworkDict()
        defers = []
        for url, url_filter in self.urls:
            defers.append(getPage(url.encode('utf8')).addCallback(self.got_bans,
                                                                  url_filter))
        DeferredList(defers).addCallback(self.bans_finished)

    def got_bans(self, data, name_filter):
        bans = json.loads(data)
        for entry in bans:
            name = entry.get('name', None)
            if name is not None and name in name_filter:
                continue
            self.new_bans[str(entry['ip'])] = str(entry['reason'])

    def bans_finished(self, _result):
        self.bans = self.new_bans
        self.new_bans = None
        log.info("successfully updated bans from bansubscribe urls")

    def get_ban(self, ip):
        if self.bans is None:
            return None
        try:
            return self.bans[ip]
        except KeyError:
            return None



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import math


def vector_collision(vec1, vec2, distance=3):
    return (math.fabs(vec1.x - vec2.x) < distance and
            math.fabs(vec1.y - vec2.y) < distance and
            math.fabs(vec1.z - vec2.z) < distance)


def collision_3d(x1, y1, z1, x2, y2, z2, distance=3):
    return (math.fabs(x1 - x2) < distance and
            math.fabs(y1 - y2) < distance and
            math.fabs(z1 - z2) < distance)


def distance_3d_vector(vector1, vector2):
    xd = vector1.x - vector2.x
    yd = vector1.y - vector2.y
    zd = vector1.z - vector2.z
    return math.sqrt(xd**2 + yd**2 + zd**2)


def distance_3d(xyz1, xyz2):
    (x1, y1, z1) = xyz1
    (x2, y2, z2) = xyz2
    xd = x1 - x2
    yd = y1 - y2
    zd = z1 - z2
    return math.sqrt(xd**2 + yd**2 + zd**2)



from pyspades.constants import NEUTRAL_TEAM, TC_CAPTURE_RATE, SPAWN_RADIUS
from pyspades.common import Vertex3
from pyspades import contained as loaders

from twisted.internet import reactor

move_object = loaders.MoveObject()
progress_bar = loaders.ProgressBar()
territory_capture = loaders.TerritoryCapture()

class Entity(Vertex3):
    team = None

    def __init__(self, entity_id, protocol, *arg, **kw):
        Vertex3.__init__(self, *arg, **kw)
        self.id = entity_id
        self.protocol = protocol

    def update(self):
        move_object.object_type = self.id
        if self.team is None:
            state = NEUTRAL_TEAM
        else:
            state = self.team.id
        move_object.state = state
        move_object.x = self.x
        move_object.y = self.y
        move_object.z = self.z
        self.protocol.send_contained(move_object, save=True)


class Flag(Entity):
    player = None

    def update(self):
        if self.player is not None:
            return
        Entity.update(self)


class Territory(Flag):
    progress = 0.0
    players = None
    start = None
    rate = 0
    rate_value = 0.0
    finish_call = None
    capturing_team = None

    def __init__(self, *arg, **kw):
        Flag.__init__(self, *arg, **kw)
        self.players = set()

    def add_player(self, player):
        self.get_progress(True)
        self.players.add(player)
        self.update_rate()

    def remove_player(self, player):
        self.get_progress(True)
        self.players.discard(player)
        self.update_rate()

    def update_rate(self):
        rate = 0
        for player in self.players:
            if player.team.id:
                rate += 1
            else:
                rate -= 1
        progress = self.progress
        if ((progress == 1.0 and (rate > 0 or rate == 0)) or
                (progress == 0.0 and (rate < 0 or rate == 0))):
            return
        self.rate = rate
        self.rate_value = rate * TC_CAPTURE_RATE
        if self.finish_call is not None:
            self.finish_call.cancel()
            self.finish_call = None
        if rate != 0:
            self.start = reactor.seconds()
            rate_value = self.rate_value
            if rate_value < 0:
                self.capturing_team = self.protocol.blue_team
                end_time = progress / -rate_value
            else:
                self.capturing_team = self.protocol.green_team
                end_time = (1.0 - progress) / rate_value
            if self.capturing_team is not self.team:
                self.finish_call = reactor.callLater(end_time, self.finish)
        self.send_progress()

    def send_progress(self):
        progress_bar.object_index = self.id
        if self.team is None:
            capturing_team = self.capturing_team
            team = capturing_team.other
        else:
            capturing_team = self.team.other
            team = self.team
        progress_bar.capturing_team = capturing_team.id
        rate = self.rate
        progress = self.get_progress()
        if team.id:
            rate = -rate
            progress = 1 - progress
        progress_bar.progress = progress
        progress_bar.rate = rate
        self.protocol.send_contained(progress_bar)

    def finish(self):
        self.finish_call = None
        protocol = self.protocol
        if self.rate > 0:
            team = protocol.green_team
        else:
            team = protocol.blue_team
        team.score += 1
        if self.team is not None:
            self.team.score -= 1
        self.team = team
        protocol.on_cp_capture(self)
        if team.score >= protocol.max_score:
            protocol.reset_game(territory=self)
            protocol.on_game_end()
        else:
            territory_capture.object_index = self.id
            territory_capture.state = self.team.id
            territory_capture.winning = False
            protocol.send_contained(territory_capture)

    def get_progress(self, set=False):
        """
        Return progress (between 0 and 1 - 0 is full blue control,
        1 is full green control) and optionally set the current
        progress.
        """
        # TODO: wtf is this thing, and why are we setting values in
        # a getter. Should this be refactored into two functions
        # called `update_progress` and `get_progress` instead?
        rate = self.rate_value
        if rate == 0.0 or self.start is None:
            return self.progress
        dt = reactor.seconds() - self.start
        progress = max(0, min(1, self.progress + rate * dt))
        if set:
            self.progress = progress
        return progress

    def get_spawn_location(self):
        x1 = max(0, self.x - SPAWN_RADIUS)
        y1 = max(0, self.y - SPAWN_RADIUS)
        x2 = min(512, self.x + SPAWN_RADIUS)
        y2 = min(512, self.y + SPAWN_RADIUS)
        return self.protocol.get_random_location(True, (x1, y1, x2, y2))


class Base(Entity):
    pass



"""
A tool for identifying griefers.

.. note::
  "blockinfo" must be AFTER "votekick" in the config script list

Commands
^^^^^^^^

* ``/griefcheck or /gc <player> <minutes>`` gives you when, how many and whos blocks a player destroyed *admin only*

Options
^^^^^^^

.. code-block:: guess

   [blockinfo]
   griefcheck_on_votekick = true
   irc_only = false

.. codeauthor:: hompy
"""

from twisted.internet.reactor import seconds
from pyspades.collision import distance_3d_vector
from pyspades.common import prettify_timespan
from piqueserver.commands import command, admin, get_player
from piqueserver.config import config

blockinfo_config = config.section("blockinfo")
GRIEFCHECK_ON_VOTEKICK = blockinfo_config.option("griefcheck_on_votekick", True)
IRC_ONLY = blockinfo_config.option("irc_only", False)


@command('griefcheck', 'gc')
def grief_check(connection, player, minutes=2):
    player = get_player(connection.protocol, player)
    protocol = connection.protocol
    color = connection not in protocol.players.values() and connection.colors
    minutes = float(minutes)
    if minutes <= 0.0:
        raise ValueError('minutes must be number greater than 0')
    time = seconds() - minutes * 60.0
    blocks_removed = player.blocks_removed or []
    blocks = [b[1] for b in blocks_removed if b[0] >= time]
    player_name = player.name
    if color:
        player_name = (('\x0303' if player.team.id else '\x0302') +
                       player_name + '\x0f')
    message = '%s removed %s block%s in the last ' % (
        player_name, len(blocks) or 'no', '' if len(blocks) == 1 else 's')
    if minutes == 1.0:
        minutes_s = 'minute'
    else:
        minutes_s = '{:.1f} minutes'.format(minutes)
    message += minutes_s + '.'
    if len(blocks):
        infos = set(blocks)
        infos.discard(None)
        if color:
            names = [('\x0303' if team else '\x0302') + name for name, team in
                     infos]
        else:
            names = set([name for name, team in infos])
        if len(names) > 0:
            message += (' Some of them were placed by ' +
                        ('\x0f, ' if color else ', ').join(names))
            message += '\x0f.' if color else '.'
        else:
            message += ' All of them were map blocks.'
        last = blocks_removed[-1]
        time_s = prettify_timespan(seconds() - last[0], get_seconds=True)
        message += ' Last one was destroyed %s ago' % time_s
        whom = last[1]
        if whom is None and len(names) > 0:
            message += ', and was part of the map'
        elif whom is not None:
            name, team = whom
            if color:
                name = ('\x0303' if team else '\x0302') + name + '\x0f'
            message += ', and belonged to %s' % name
        message += '.'
    switch_sentence = False
    if player.last_switch is not None and player.last_switch >= time:
        time_s = prettify_timespan(seconds() - player.last_switch,
                                   get_seconds=True)
        message += ' %s joined %s team %s ago' % (player_name,
                                                  player.team.name, time_s)
        switch_sentence = True
    teamkills = len([t for t in player.teamkill_times or [] if t >= time])
    if teamkills > 0:
        s = ', and killed' if switch_sentence else ' %s killed' % player_name
        message += s + ' %s teammates in the last %s' % (teamkills, minutes_s)
    if switch_sentence or teamkills > 0:
        message += '.'
    votekick = getattr(protocol, 'votekick', None)
    if (votekick and votekick.victim is player and
            votekick.victim.world_object and votekick.instigator.world_object):
        instigator = votekick.instigator
        tiles = int(distance_3d_vector(player.world_object.position,
                                       instigator.world_object.position))
        instigator_name = (('\x0303' if instigator.team.id else '\x0302') +
                           instigator.name + '\x0f')
        message += (' %s is %d tiles away from %s, who started the votekick.' %
                    (player_name, tiles, instigator_name))
    return message


def apply_script(protocol, connection, config):

    class BlockInfoConnection(connection):
        blocks_removed = None
        teamkill_times = None

        def on_reset(self):
            self.blocks_removed = None
            self.teamkill_times = None
            connection.on_reset(self)

        def on_block_build(self, x, y, z):
            if self.protocol.block_info is None:
                self.protocol.block_info = {}
            self.protocol.block_info[(x, y, z)] = (self.name, self.team.id)
            connection.on_block_build(self, x, y, z)

        def on_line_build(self, points):
            if self.protocol.block_info is None:
                self.protocol.block_info = {}
            name_team = (self.name, self.team.id)
            for point in points:
                self.protocol.block_info[point] = name_team
            connection.on_line_build(self, points)

        def on_block_removed(self, x, y, z):
            if self.protocol.block_info is None:
                self.protocol.block_info = {}
            if self.blocks_removed is None:
                self.blocks_removed = []
            pos = (x, y, z)
            info = (seconds(), self.protocol.block_info.pop(pos, None))
            self.blocks_removed.append(info)
            connection.on_block_removed(self, x, y, z)

        def on_kill(self, killer, type, grenade):
            if killer and killer.team is self.team:
                if killer.teamkill_times is None:
                    killer.teamkill_times = []
                killer.teamkill_times.append(seconds())
            return connection.on_kill(self, killer, type, grenade)

    class BlockInfoProtocol(protocol):
        block_info = None

        def on_map_change(self, map):
            self.block_info = None
            protocol.on_map_change(self, map)

        def on_votekick_start(self, instigator, victim, reason):
            result = protocol.on_votekick_start(
                self, instigator, victim, reason)
            if result is None and GRIEFCHECK_ON_VOTEKICK.get():
                message = grief_check(instigator, victim.name)
                if IRC_ONLY.get():
                    self.irc_say('* ' + message)
                else:
                    self.send_chat(message, irc=True)
            return result

    return BlockInfoProtocol, BlockInfoConnection



"""
Lets you set restrictions on spectators.

Original documentation:

This script will hopefully give server owners some control over what
spectators do on there server. As of now since the release of v0.75,
Goon Haven has had issues with spectators idling and using global chat
to send information to a team so that they may know enemy positions
or what the enemy is doing, etc. This script can block spectator chat
as well as kick spectators after so much time as passed.

Additionally, server owners who also give out "guard" or "mini-mod"
positions can add the right "specpower" to the group rights in commands.py
to have the guards/minimods be immune to the spectator kick and chat
restrictions.

Oh, and admins are also automatically immune to spectator kick and chat
restrictions.

Hope you enjoy!
Tocksman

Options
^^^^^^^

.. code-block:: guess

   [spectator_control]
   no_chat = false # determines whether spectators can chat or not in your server
   kick = false # determines whether spectators will be kicked after remaining for so long
   kick_time = "5min" # how long a spectator may remain before they are kicked

.. codeauthor:: Tocksman (made for Goon Haven)
"""

from math import ceil, floor
from twisted.internet import reactor
from piqueserver.config import config, cast_duration

spectator_ctrl_config = config.section("spectator_control")
no_chat = spectator_ctrl_config.option("no_chat", False)
kick = spectator_ctrl_config.option("kick", False)
kick_time = spectator_ctrl_config.option("kick_time", default="5min", cast=cast_duration)

def apply_script(protocol, connection, config):
    class SpectatorControlConnection(connection):
        spec_check = None

        def on_chat(self, value, global_message):
            # if no chat is set and they're a spectator and not an admin
            # also, check for the right "specpower" for owners who add additional
            # rights such as guards, mini-mods, etc.
            if self.team.spectator and no_chat.get():
                if not self.admin and not self.rights.specpower:  # not an admin
                    self.send_chat('Spectators cannot speak on this server.')
                    return False  # deny
            return connection.on_chat(self, value, global_message)

        def on_team_join(self, team):
            if team.spectator and kick.get() and kick_time.get() > 0:
                if self.rights is None or (not self.admin and not self.rights.specpower):  # not an admin
                    # this check is necessary as you can join spectator from
                    # being a spectator
                    if self.spec_check is None or not self.spec_check.active():
                        self.send_chat(
                            'Warning! Spectators are kicked after %s seconds!' %
                            (kick_time.get()))
                        time = ceil((kick_time.get() / 4) * 3)
                        self.spec_check = reactor.callLater(
                            time, self.check_spec_time, 1)
            elif not team.spectator:
                if self.spec_check is not None and self.spec_check.active():
                    self.spec_check.cancel()
                    self.spec_check = None
            return connection.on_team_join(self, team)

        def on_disconnect(self):
            if self.spec_check is not None and self.spec_check.active():
                self.spec_check.cancel()
            self.spec_check = None
            return connection.on_disconnect(self)

        def check_spec_time(self, id):
            if not self.team.spectator:
                print(
                    'WARNING 1. Safety check kept an non-spectator from being spectator-kicked. Report this please!')
                return
            if self.admin or self.rights.specpower:
                print(
                    'WARNING 2. Safety check kept an admin from being spectator-kicked.')
                return
            if id == 1:
                seconds = floor(kick_time.get() / 4)
                self.send_chat(
                    'Warning! If you do not leave spectator, you will be kicked in %s seconds!' %
                    (seconds))
                self.spec_check = reactor.callLater(
                    seconds, self.check_spec_time, 2)
            elif id == 2:
                self.kick(
                    'You have been kicked for remaining in spectator for too long.')
    return protocol, SpectatorControlConnection



"""
Allows users to start votekicks

Commands
^^^^^^^^

* ``/votekick <player> <reason>`` start votekick against a player
* ``/y`` votes yes
* ``/togglevotekick or /tvk`` toggles votekicks on/off globally
* ``/togglevotekick or /tvk <player>`` toggles votekicks on/off for specific players
* ``/cancel`` cancels a votekick

Options
^^^^^^^

.. code-block:: guess

    [votekick]
    # percentage of total number of players in the server required to vote to
    # successfully votekick a player
    percentage = 35

    # duration that votekicked player will be banned for
    ban_duration = "30min"

    public_votes = true

.. codeauthor:: James Hofmann a.k.a triplefox
"""


from twisted.internet.reactor import seconds
from piqueserver.scheduler import Scheduler
from piqueserver.commands import (command, admin, get_player, join_arguments,
                                  CommandError, player_only)
from piqueserver.config import config, cast_duration

REQUIRE_REASON = True

S_VOTEKICKING_DISABLED = 'Votekicking disabled'
S_VOTEKICKING_SET = 'Votekicking globally {set}.'
S_VOTEKICK_DISALLOWED = 'You are not allowed to initiate a votekick.'
S_VOTEKICK_USER_SET = 'Votekicking is {set} for {user}.'
S_NO_VOTEKICK = 'No votekick in progress'
S_DEFAULT_REASON = 'NO REASON GIVEN'
S_IN_PROGRESS = 'Votekick already in progress'
S_SELF_VOTEKICK = "You can't votekick yourself"
S_NOT_ENOUGH_PLAYERS = "There aren't enough players to vote"
S_VOTEKICK_IMMUNE = "You can't votekick this player"
S_NOT_YET = "You can't start another votekick yet!"
S_NEED_REASON = 'You must provide a reason for the votekick'
S_CANT_CANCEL = "You didn't start the votekick!"
S_YES = '{player} voted YES'
S_ENDED = 'Votekick for {victim} has ended. {result}'
S_RESULT_TIMED_OUT = 'Votekick timed out'
S_RESULT_CANCELLED = 'Cancelled'
S_RESULT_BANNED = 'Banned by admin'
S_RESULT_KICKED = 'Kicked by admin'
S_RESULT_INSTIGATOR_KICKED = 'Instigator kicked by admin'
S_RESULT_LEFT = '{victim} left during votekick'
S_RESULT_INSTIGATOR_LEFT = 'Instigator {instigator} left'
S_RESULT_PASSED = 'Player kicked'
S_ANNOUNCE_IRC = '* {instigator} started a votekick against player {victim}. ' \
    'Reason: {reason}'
S_ANNOUNCE = '{instigator} started a VOTEKICK against {victim}. Say /Y to agree'
S_ANNOUNCE_SELF = 'You started a votekick against {victim}. Say /CANCEL to ' \
    'stop it'
S_UPDATE = '{instigator} is votekicking {victim}. /Y to vote ({needed} left)'
S_REASON = 'Reason: {reason}'

# register options
VOTEKICK_CONFIG = config.section('votekick')
REQUIRED_PERCENTAGE_OPTION = VOTEKICK_CONFIG.option('percentage', 35.0)
BAN_DURATION_OPTION = VOTEKICK_CONFIG.option('ban_duration', default="30min", cast=cast_duration)
PUBLIC_VOTES_OPTION = VOTEKICK_CONFIG.option('public_votes', True)

class VotekickFailure(Exception):
    pass


@command('votekick')
@player_only
def start_votekick(connection, *args):
    """
    Starts an votekick against a player
    /votekick <player name or id> <reason>
    """
    protocol = connection.protocol
    player = connection

    if not protocol.votekick_enabled:
        return S_VOTEKICKING_DISABLED
    if not player.votekick_enabled:
        return S_VOTEKICK_DISALLOWED

    if not args:
        if protocol.votekick:
            # player requested votekick info
            protocol.votekick.send_chat_update(player)
            return
        raise ValueError("Target player is required")

    value = args[0]
    victim = get_player(protocol, value)
    reason = join_arguments(args[1:])

    try:
        # attempt to start votekick
        votekick = Votekick.start(player, victim, reason)
        protocol.votekick = votekick
    except VotekickFailure as err:
        return str(err)


@command('cancel')
def cancel_votekick(connection):
    """
    Cancels an ongoing vote
    /cancel
    """
    protocol = connection.protocol
    votekick = protocol.votekick
    if not votekick:
        return S_NO_VOTEKICK
    if connection in protocol.players.values():
        player = connection
        if (player is not votekick.instigator and not player.admin and
                not player.rights.cancel):
            return S_CANT_CANCEL

    votekick.end(S_RESULT_CANCELLED)


@command('y')
@player_only
def vote_yes(connection):
    """
    Vote yes on an ongoing vote
    /y
    """
    player = connection

    votekick = connection.protocol.votekick
    if not votekick:
        return S_NO_VOTEKICK

    votekick.vote(player)


@command('tvk', admin_only=True)
def togglevotekick(connection, *args):
    """
    Toggles votekicking for a player or the whole server
    /tvk <player> or /tvk
    """
    protocol = connection.protocol
    if len(args) == 0:
        protocol.votekick_enabled = not protocol.votekick_enabled
        return S_VOTEKICKING_SET.format(
            set=('enabled' if protocol.votekick_enabled else 'disabled'))
    try:
        player = get_player(protocol, args[0])
    except CommandError:
        return 'Invalid Player'
    player.votekick_enabled = not player.votekick_enabled
    return S_VOTEKICK_USER_SET.format(user=player.name, set=(
        'enabled' if player.votekick_enabled else 'disabled'))


class Votekick:
    timeout = 120.0  # 2 minutes
    interval = 120.0  # 2 minutes
    ban_duration = BAN_DURATION_OPTION.get()
    public_votes = PUBLIC_VOTES_OPTION.get()
    schedule = None

    @property
    def votes_remaining(self) -> int:
        return self.protocol.get_required_votes() - len(self.votes) + 1

    @classmethod
    def start(cls, instigator, victim, reason=None):
        protocol = instigator.protocol
        last_votekick = instigator.last_votekick
        reason = reason.strip() if reason else None
        if protocol.votekick:
            raise VotekickFailure(S_IN_PROGRESS)
        elif instigator is victim:
            raise VotekickFailure(S_SELF_VOTEKICK)
        elif protocol.get_required_votes() <= 0:
            raise VotekickFailure(S_NOT_ENOUGH_PLAYERS)
        elif victim.admin or victim.rights.cancel or victim.local:
            raise VotekickFailure(S_VOTEKICK_IMMUNE)
        elif not instigator.admin and (last_votekick is not None and
                                       seconds() - last_votekick < cls.interval):
            raise VotekickFailure(S_NOT_YET)
        elif REQUIRE_REASON and not reason:
            raise VotekickFailure(S_NEED_REASON)

        result = protocol.on_votekick_start(instigator, victim, reason)
        if result is not None:
            raise VotekickFailure(result)

        reason = reason or S_DEFAULT_REASON
        return cls(instigator, victim, reason)

    def __init__(self, instigator, victim, reason):
        self.protocol = protocol = instigator.protocol
        self.instigator = instigator
        self.victim = victim
        self.reason = reason
        self.votes = {instigator: True}
        self.ended = False

        protocol.irc_say(
            S_ANNOUNCE_IRC.format(
                instigator=instigator.name,
                victim=victim.name,
                reason=self.reason))
        protocol.send_chat(
            S_ANNOUNCE.format(
                instigator=instigator.name,
                victim=victim.name),
            sender=instigator)
        protocol.send_chat(S_REASON.format(reason=self.reason),
                           sender=instigator)
        instigator.send_chat(S_ANNOUNCE_SELF.format(victim=victim.name))

        schedule = Scheduler(protocol)
        schedule.call_later(self.timeout, self.end, S_RESULT_TIMED_OUT)
        schedule.loop_call(30.0, self.send_chat_update)
        self.schedule = schedule

    def vote(self, player):
        if self.victim is player:
            return
        elif player in self.votes:
            return
        if self.public_votes:
            self.protocol.send_chat(S_YES.format(player=player.name))
        self.votes[player] = True
        if self.votes_remaining <= 0:
            # vote passed, ban or kick accordingly
            victim = self.victim
            self.end(S_RESULT_PASSED)
            print(victim.name, 'votekicked')
            if self.ban_duration > 0.0:
                victim.ban(self.reason, self.ban_duration)
            else:
                victim.kick(silent=True)

    def release(self):
        self.instigator = None
        self.victim = None
        self.votes = None
        if self.schedule:
            self.schedule.reset()
        self.schedule = None
        self.protocol.votekick = None

    def end(self, result):
        self.ended = True
        message = S_ENDED.format(victim=self.victim.name, result=result)
        self.protocol.send_chat(message, irc=True)
        if not self.instigator.admin:
            self.instigator.last_votekick = seconds()
        self.protocol.on_votekick_end()
        self.release()

    def send_chat_update(self, target=None):
        # send only to target player if provided, otherwise broadcast to server
        target = target or self.protocol
        target.send_chat(
            S_UPDATE.format(
                instigator=self.instigator.name,
                victim=self.victim.name,
                needed=self.votes_remaining))
        target.send_chat(S_REASON.format(reason=self.reason))


def apply_script(protocol, connection, config):

    class VotekickProtocol(protocol):
        votekick = None
        votekick_enabled = True

        def get_required_votes(self):
            # votekicks are invalid if this returns <= 0
            player_count = sum(not player.disconnected and not player.local
                               for player in self.players.values()) - 1
            return int(player_count / 100.0 * REQUIRED_PERCENTAGE_OPTION.get())

        def on_map_leave(self):
            if self.votekick:
                self.votekick.release()
            protocol.on_map_leave(self)

        def on_ban(self, banee, reason, duration):
            votekick = self.votekick
            if votekick and votekick.victim is self:
                votekick.end(S_RESULT_BANNED)
            protocol.on_ban(self, connection, reason, duration)

        def on_votekick_start(self, instigator, victim, reason):
            pass

        def on_votekick_end(self):
            pass

    class VotekickConnection(connection):
        last_votekick = None
        votekick_enabled = True

        def on_disconnect(self):
            votekick = self.protocol.votekick
            if votekick:
                if votekick.victim is self:
                    # victim leaves, gets votekick ban
                    reason = votekick.reason
                    votekick.end(S_RESULT_LEFT.format(victim=self.name))
                    self.ban(reason, Votekick.ban_duration)
                elif votekick.instigator is self:
                    # instigator leaves, votekick is called off
                    s = S_RESULT_INSTIGATOR_LEFT.format(instigator=self.name)
                    votekick.end(s)
                else:
                    # make sure we still have enough players
                    votekick.votes.pop(self, None)
                    if votekick.votes_remaining <= 0:
                        votekick.end(S_NOT_ENOUGH_PLAYERS)
            connection.on_disconnect(self)

        def kick(self, reason=None, silent=False):
            votekick = self.protocol.votekick
            if votekick:
                if votekick.victim is self:
                    votekick.end(S_RESULT_KICKED)
                elif votekick.instigator is self:
                    votekick.end(S_RESULT_INSTIGATOR_KICKED)
            connection.kick(self, reason, silent)

    return VotekickProtocol, VotekickConnection



from math import atan2, sqrt, pi

from piqueserver.commands import command
from pyspades.contained import BlockAction, SetColor
from pyspades.constants import BUILD_BLOCK, DESTROY_BLOCK

EAST = 0
SOUTH = 1
WEST = 2
NORTH = 3


def make_color(r, g, b, a):
    return b | (g << 8) | (r << 16) | (int((a / 255.0) * 128) << 24)


def make_color_tuple(color):
    return make_color(color[0], color[1], color[2], 255)


def get_color_tuple(color):
    b = color & 0xFF
    g = (color & 0xFF00) >> 8
    r = (color & 0xFF0000) >> 16
    a = int((((color & 0xFF000000) >> 24) / 128.0) * 255)
    return (r, g, b, a)


def set_color(prt, color, player_id=32):
    c = SetColor()
    c.player_id = player_id
    c.value = color
    prt.broadcast_contained(c)


def add_block(
        prt,
        x,
        y,
        z,
        color,
        player_id=32,
        mirror_x=False,
        mirror_y=False):
    if x >= 0 and x < 512 and y >= 0 and y < 512 and z >= 0 and z < 64:
        if mirror_x or mirror_y:
            x2 = x
            y2 = y
            if mirror_x:
                x2 = 511 - x
            if mirror_y:
                y2 = 511 - y
            add_block(prt, x2, y2, z, color, player_id, False, False)
        if not prt.map.get_solid(x, y, z):
            block_action = BlockAction()
            block_action.player_id = player_id
            block_action.value = BUILD_BLOCK
            block_action.x = x
            block_action.y = y
            block_action.z = z
            prt.broadcast_contained(block_action)
            prt.map.set_point(x, y, z, get_color_tuple(color))


def remove_block(prt, x, y, z, mirror_x=False, mirror_y=False):
    if x >= 0 and x < 512 and y >= 0 and y < 512 and z >= 0 and z < 64:
        if mirror_x or mirror_y:
            x2 = x
            y2 = y
            if mirror_x:
                x2 = 511 - x
            if mirror_y:
                y2 = 511 - y
            remove_block(prt, x2, y2, z, False, False)
        if prt.map.get_solid(x, y, z):
            block_action = BlockAction()
            block_action.player_id = 32
            block_action.value = DESTROY_BLOCK
            block_action.x = x
            block_action.y = y
            block_action.z = z
            prt.map.remove_point(x, y, z)
            prt.broadcast_contained(block_action)
            return True
    return False


@command()
def mirror(connection, mirror_x, mirror_y):
    connection.mirror_x = bool(mirror_x)
    connection.mirror_y = bool(mirror_y)


@command()
def tunnel(connection, radius, length, zoffset=0):
    connection.reset_build()
    connection.callback = tunnel_r
    connection.arguments = [connection, radius, length, zoffset]
    connection.select = True
    connection.points = 1


def tunnel_r(connection, radius, length, zoffset=0):
    radius = int(radius)
    length = int(length)
    zoffset = int(zoffset)
    facing = connection.get_direction()
    if facing == WEST or facing == NORTH:
        length = -length
    for rel_h in range(-radius, radius + 1):
        for rel_v in range(-radius, 1):
            if round(sqrt(rel_h**2 + rel_v**2)) <= radius:
                if facing == NORTH or facing == SOUTH:
                    y1 = connection.block1_y
                    y2 = connection.block1_y + length
                    for y in range(min(y1, y2), max(y1, y2) + 1):
                        remove_block(
                            connection.protocol,
                            connection.block1_x +
                            rel_h,
                            y,
                            connection.block1_z +
                            rel_v +
                            zoffset,
                            connection.mirror_x,
                            connection.mirror_y)
                elif facing == WEST or facing == EAST:
                    x1 = connection.block1_x
                    x2 = connection.block1_x + length
                    for x in range(min(x1, x2), max(x1, x2) + 1):
                        remove_block(
                            connection.protocol,
                            x,
                            connection.block1_y +
                            rel_h,
                            connection.block1_z +
                            rel_v +
                            zoffset,
                            connection.mirror_x,
                            connection.mirror_y)


@command()
def insert(connection):
    connection.reset_build()
    connection.callback = insert_r
    connection.arguments = [connection]
    connection.select = True
    connection.points = 2


def insert_r(connection):
    x1 = min(connection.block1_x, connection.block2_x)
    x2 = max(connection.block1_x, connection.block2_x)
    y1 = min(connection.block1_y, connection.block2_y)
    y2 = max(connection.block1_y, connection.block2_y)
    z1 = min(connection.block1_z, connection.block2_z)
    z2 = max(connection.block1_z, connection.block2_z)
    color = make_color_tuple(connection.color)
    for xx in range(x1, x2 + 1):
        for yy in range(y1, y2 + 1):
            for zz in range(z1, z2 + 1):
                add_block(
                    connection.protocol,
                    xx,
                    yy,
                    zz,
                    color,
                    connection.player_id,
                    connection.mirror_x,
                    connection.mirror_y)


@command()
def delete(connection):
    connection.reset_build()
    connection.callback = delete_r
    connection.arguments = [connection]
    connection.select = True
    connection.points = 2


def delete_r(connection):
    x1 = min(connection.block1_x, connection.block2_x)
    x2 = max(connection.block1_x, connection.block2_x)
    y1 = min(connection.block1_y, connection.block2_y)
    y2 = max(connection.block1_y, connection.block2_y)
    z1 = min(connection.block1_z, connection.block2_z)
    z2 = max(connection.block1_z, connection.block2_z)
    for xx in range(x1, x2 + 1):
        for yy in range(y1, y2 + 1):
            for zz in range(z1, z2 + 1):
                remove_block(connection.protocol, xx, yy, zz,
                             connection.mirror_x, connection.mirror_y)


@command()
def pattern(connection, copies=1):
    connection.reset_build()
    connection.callback = pattern_r
    connection.arguments = [connection, copies]
    connection.select = True
    connection.points = 2


def pattern_r(connection, copies=1):
    copies = int(copies)
    x1 = min(connection.block1_x, connection.block2_x)
    x2 = max(connection.block1_x, connection.block2_x)
    y1 = min(connection.block1_y, connection.block2_y)
    y2 = max(connection.block1_y, connection.block2_y)
    z1 = min(connection.block1_z, connection.block2_z)
    z2 = max(connection.block1_z, connection.block2_z)
    delta_z = (z2 - z1) + 1
    for xx in range(x1, x2 + 1):
        for yy in range(y1, y2 + 1):
            for zz in range(z1, z2 + 1):
                if connection.protocol.map.get_solid(xx, yy, zz):
                    color = make_color_tuple(
                        connection.protocol.map.get_point(xx, yy, zz)[1])
                    set_color(connection.protocol, color, 32)
                    for i in range(1, copies + 1):
                        z_offset = delta_z * i
                        add_block(
                            connection.protocol,
                            xx,
                            yy,
                            zz - z_offset,
                            color,
                            32,
                            connection.mirror_x,
                            connection.mirror_y)


@command()
def hollow(connection, thickness=1):
    connection.reset_build()
    connection.callback = hollow_r
    connection.arguments = [connection, thickness]
    connection.select = True
    connection.points = 2


def hollow_r(connection, thickness=1):
    m = connection.protocol.map
    thickness = int(thickness) - 1
    x1 = min(connection.block1_x, connection.block2_x)
    x2 = max(connection.block1_x, connection.block2_x)
    y1 = min(connection.block1_y, connection.block2_y)
    y2 = max(connection.block1_y, connection.block2_y)
    z1 = min(connection.block1_z, connection.block2_z)
    z2 = max(connection.block1_z, connection.block2_z)
    blocks = []
    xr = x2 - x1 + 1
    yr = y2 - y1 + 1
    zr = z2 - z1 + 1
    for x in range(0, xr):
        blocks.append([])
        for y in range(0, yr):
            blocks[x].append([])
            for z in range(0, zr):
                blocks[x][y].append(False)

    def hollow_check(xc, yc, zc, thickness):
        if thickness > 0:
            for xx in range(xc - 1, xc + 2):
                if xx >= 0 and xx < xr:
                    for yy in range(yc - 1, yc + 2):
                        if yy >= 0 and yy < yr:
                            for zz in range(zc - 1, zc + 2):
                                if zz >= 0 and zz < zr:
                                    blocks[xx][yy][zz] = True
                                    if m.get_solid(x1 + xx, y1 + yy, z1 + zz):
                                        hollow_check(xx, yy, zz, thickness - 1)
    for x in range(0, xr):
        for y in range(0, yr):
            for z in range(0, zr):
                if m.get_solid(x1 + x, y1 + y, z1 + z):
                    if m.is_surface(x1 + x, y1 + y, z1 + z):
                        blocks[x][y][z] = True
                        hollow_check(x, y, z, thickness)
                else:
                    blocks[x][y][z] = True
    for x in range(0, xr):
        for y in range(0, yr):
            for z in range(0, zr):
                if not blocks[x][y][z]:
                    remove_block(connection.protocol, x1 + x, y1 + y, z1 + z)


def apply_script(protocol, connection, config):
    class MapMakingToolsConnection(connection):
        select = False
        mirror_x = False
        mirror_y = False

        def reset_build(self):
            self.block1_x = None
            self.block1_y = None
            self.block1_z = None
            self.block2_x = None
            self.block2_y = None
            self.block2_z = None
            self.callback = None
            self.arguments = None
            self.select = False
            self.points = None

        def get_direction(self):
            orientation = self.world_object.orientation
            angle = atan2(orientation.y, orientation.x)
            if angle < 0:
                angle += pi*2
            # Convert to units of quadrents
            angle *= 2/pi
            angle = round(angle)
            if angle == 4:
                angle = 0
            return angle

        def on_block_destroy(self, x, y, z, value):
            if self.select:
                if self.points == 1:
                    self.block1_x = x
                    self.block1_y = y
                    self.block1_z = z
                    try:
                        self.callback(*self.arguments)
                    except Exception as e:
                        self.send_chat('Map building failed: {}'.format(e))
                    self.reset_build()
                    return False
                elif self.points == 2:
                    if self.block1_x is None:
                        self.block1_x = x
                        self.block1_y = y
                        self.block1_z = z
                        self.send_chat('First block selected')
                        return False
                    else:
                        if not (x == self.block1_x and y ==
                                self.block1_y and z == self.block1_z):
                            self.block2_x = x
                            self.block2_y = y
                            self.block2_z = z
                            try:
                                self.callback(*self.arguments)
                            except Exception as e:
                                self.send_chat('Map building failed: {}'.format(e))
                            self.reset_build()
                            self.send_chat('Second block selected')
                            return False
            if self.mirror_x or self.mirror_y:
                x2 = x
                y2 = y
                if self.mirror_x:
                    x2 = 511 - x
                if self.mirror_y:
                    y2 = 511 - y
                remove_block(self.protocol, x2, y2, z)
            connection.on_block_destroy(self, x, y, z, value)

        def on_block_build(self, x, y, z):
            if self.mirror_x or self.mirror_y:
                x2 = x
                y2 = y
                if self.mirror_x:
                    x2 = 511 - x
                if self.mirror_y:
                    y2 = 511 - y
                add_block(self.protocol, x2, y2, z,
                          make_color_tuple(self.color), self.player_id)
            connection.on_block_build(self, x, y, z)

    return protocol, MapMakingToolsConnection



import math
import re
from twisted.internet import reactor
from piqueserver.config import cast_duration
from pyspades.common import prettify_timespan
from piqueserver.commands import (
    command, CommandError, get_player, get_team, get_truthy, target_player)


@command('time')
def get_time_limit(connection):
    """
    Tell you the current time limit
    /time
    """
    advance_call = connection.protocol.advance_call
    if advance_call is None:
        return 'No time limit set'
    left = int(
        math.ceil((advance_call.getTime() - reactor.seconds()) / 60.0))
    return 'There are %s minutes left' % left


@command("resetgame", admin_only=True)
def reset_game(connection):
    """
    Reset the game
    /resetgame
    """
    resetting_player = connection
    # irc compatibility
    if resetting_player not in connection.protocol.players.values():
        for player in connection.protocol.players.values():
            resetting_player = player
            if player.admin:
                break
        if resetting_player is connection:
            return
    connection.protocol.reset_game(resetting_player)
    connection.protocol.on_game_end()
    connection.protocol.send_chat(
        'Game has been reset by %s' % connection.name,
        irc=True)


@command(admin_only=True)
def lock(connection, value):
    """
    Make a specified team no longer joinable until it's unlocked
    /lock <blue|green|spectator>
    New players will be placed in the spectator team even if it's locked
    """
    team = get_team(connection, value)
    team.locked = True
    connection.protocol.send_chat('%s team is now locked' % team.name)
    connection.protocol.irc_say('* %s locked %s team' % (connection.name,
                                                         team.name))


@command(admin_only=True)
def unlock(connection, value):
    """
    Unlock a team
    /unlock <blue|green|spectator>
    """
    team = get_team(connection, value)
    team.locked = False
    connection.protocol.send_chat('%s team is now unlocked' % team.name)
    connection.protocol.irc_say('* %s unlocked %s team' % (connection.name,
                                                           team.name))


@command(admin_only=True)
@target_player
def switch(connection, player):
    """
    Switch teams either for yourself or for a given player
    /switch [player]
    """
    protocol = connection.protocol
    if player.team.spectator:
        player.send_chat(
            "The switch command can't be used on a spectating player.")
        return
    new_team = player.team.other
    if player.invisible:
        old_team = player.team
        player.team = new_team
        player.on_team_changed(old_team)
        player.spawn(player.world_object.position.get())
        player.send_chat('Switched to %s team' % player.team.name)
        if connection is not player and connection in protocol.players.values():
            connection.send_chat('Switched %s to %s team' % (player.name,
                                                             player.team.name))
        protocol.irc_say('* %s silently switched teams' % player.name)
    else:
        player.respawn_time = protocol.respawn_time
        player.set_team(new_team)
        protocol.send_chat('%s switched teams' % player.name, irc=True)


@command('setbalance', admin_only=True)
def set_balance(connection, value):
    """
    Turn automatic balancing on or off
    /setbalance <on|off>
    """
    should_balance = get_truthy(value)
    if should_balance is None:
        raise CommandError()

    protocol = connection.protocol
    protocol.balanced_teams = should_balance

    if should_balance:
        protocol.send_chat('now balancing teams')
        connection.protocol.irc_say(
            '* %s turned on balanced teams' % connection.name)
    else:
        protocol.send_chat('now no longer balancing teams')
        connection.protocol.irc_say(
            '* %s turned off balanced teams' % connection.name)


@command('togglebuild', 'tb', admin_only=True)
def toggle_build(connection, player=None):
    """
    Toggle building for everyone in the server or for a given player
    /togglebuild [player]
    """
    if player is not None:
        player = get_player(connection.protocol, player)
        value = not player.building
        player.building = value
        msg = '%s can build again' if value else '%s is disabled from building'
        connection.protocol.send_chat(msg % player.name)
        connection.protocol.irc_say('* %s %s building for %s' %
                                    (connection.name,
                                     ['disabled', 'enabled'][int(value)],
                                     player.name))
    else:
        value = not connection.protocol.building
        connection.protocol.building = value
        on_off = ['OFF', 'ON'][int(value)]
        connection.protocol.send_chat('Building has been toggled %s!' % on_off)
        connection.protocol.irc_say(
            '* %s toggled building %s' % (connection.name,
                                          on_off))


@command('togglekill', 'tk', admin_only=True)
def toggle_kill(connection, player=None):
    """
    Toggle killing for everyone in the server or for a given player
    /togglekill [player]
    """
    if player is not None:
        player = get_player(connection.protocol, player)
        value = not player.killing
        player.killing = value
        msg = '%s can kill again' if value else '%s is disabled from killing'
        connection.protocol.send_chat(msg % player.name)
        connection.protocol.irc_say('* %s %s killing for %s' %
                                    (connection.name,
                                     ['disabled', 'enabled'][int(value)], player.name))
    else:
        value = not connection.protocol.killing
        connection.protocol.killing = value
        on_off = ['OFF', 'ON'][int(value)]
        connection.protocol.send_chat('Killing has been toggled %s!' % on_off)
        connection.protocol.irc_say(
            '* %s toggled killing %s' % (connection.name,
                                         on_off))


@command('toggleteamkill', 'ttk', admin_only=True)
def toggle_teamkill(connection):
    """
    Toggle friendly fire
    /toggleteamkill
    """
    value = not connection.protocol.friendly_fire
    connection.protocol.friendly_fire = value
    on_off = ['OFF', 'ON'][int(value)]
    connection.protocol.send_chat(
        'Friendly fire has been toggled %s!' % on_off)
    connection.protocol.irc_say('* %s toggled friendly fire %s' % (
        connection.name, on_off))


@command('globalchat', admin_only=True)
def global_chat(connection, value=None):
    """
    Enable or disable global chat
    /globalchat [on|off]
    Toggles if no arguments are given
    """
    enabled = get_truthy(value)
    if enabled is True:
        connection.protocol.global_chat = True
    elif enabled is False:
        connection.protocol.global_chat = False
    else:
        connection.protocol.global_chat = not connection.protocol.global_chat

    connection.protocol.send_chat(
        'Global chat %s' % (
            'enabled' if connection.protocol.global_chat else 'disabled'),
        irc=True)


@command('timelimit', admin_only=True)
def set_time_limit(connection, duration):
    """
    Set this game time limit
    /timelimit <duration>
    """
    limit = cast_duration(duration)
    span = prettify_timespan(limit)
    protocol = connection.protocol
    # takes time in minutes
    protocol.set_time_limit(limit/60)
    protocol.send_chat('Time limit set to {}'.format(span), irc=True)


@command(admin_only=True)
def fog(connection, *args):
    """
    Set the fog color
    /fog red green blue (all values 0-255)
    /fog #aabbcc        (hex representation of rgb)
    /fog #abc           (short hex representation of rgb)
    """
    if(len(args) == 3):
        r = int(args[0])
        g = int(args[1])
        b = int(args[2])
    elif(len(args) == 1 and args[0][0] == '#'):
        hex_code = args[0][1:]

        if (len(hex_code) != 3) and (len(hex_code) != 6):
            raise ValueError("Invalid hex code length")

        if len(hex_code) == 3:  # it's a short hex code, turn it into a full one
            hex_code = (hex_code[0]*2) + (hex_code[1]*2) + (hex_code[2]*2)

        valid_characters = re.compile('[a-fA-F0-9]')
        for char in hex_code:
            if valid_characters.match(char) is None:
                raise ValueError("Invalid hex code characters")

        r = int(hex_code[0:2], base=16)
        g = int(hex_code[2:4], base=16)
        b = int(hex_code[4:6], base=16)
    else:
        raise ValueError("Neither RGB or hex code provided")

    old_fog_color = connection.protocol.fog_color
    connection.protocol.set_fog_color((r, g, b))
    if old_fog_color == (r, g, b):
        return 'Fog color changed successfully\nWarning: fog color set to same color as before'
    return 'Fog color changed successfully'



from pyspades.common import (coordinates, to_coordinates)
from piqueserver.commands import (command, CommandError, get_player,
                                  PermissionDenied, player_only, target_player)


@command(admin_only=True)
@target_player
def unstick(connection, player):
    """
    Unstick yourself or another player and inform everyone on the server of it
    /unstick [player]
    """
    connection.protocol.send_chat("%s unstuck %s" %
                                  (connection.name, player.name), irc=True)
    player.set_location_safe(player.get_location())


@command('moves', admin_only=True)
def move_silent(connection, *args):
    """
    Silently move yourself or a given player to the specified x/y/z coordinates or sector
    /moves [player] <sector> or /moves [player] <x> <y> <z>
    If the z coordinate makes the player appear underground, put them at ground level instead.
    If the x/y/z coordinate makes the player appear outside of the world bounds,
    take the bound instead

    You can only move other players if you are admin or have the move_others right
    """
    do_move(connection, args, True)


@command(admin_only=True)
@player_only
def move(connection, *args):
    """
    Move yourself or a given player to the specified x/y/z coordinates or sector
    /move [player] <sector> or /move [player] <x> <y> <z>
    If you're invisible, it will happen silently.
    If the z coordinate makes the player appear underground, put them at ground level instead.
    If the x/y/z coordinate makes the player appear outside of the world bounds,
    take the bound instead

    You can only move other players if you are admin or have the move_others right
    """
    do_move(connection, args)


def do_move(connection, args, silent=False):
    position = None
    player = None
    arg_count = len(args)

    initial_index = 1 if arg_count == 2 or arg_count == 4 else 0

    # the target position is a <sector>
    if arg_count == 1 or arg_count == 2:
        x, y = coordinates(args[initial_index])
        x += 32
        y += 32
        z = connection.protocol.map.get_height(x, y) - 2
        position = args[initial_index].upper()
    # the target position is <x> <y> <z>
    elif arg_count == 3 or arg_count == 4:
        x = min(max(0, int(args[initial_index])), 511)
        y = min(max(0, int(args[initial_index + 1])), 511)
        z = min(max(0, int(args[initial_index + 2])),
                connection.protocol.map.get_height(x, y) - 2)
        position = '%d %d %d' % (x, y, z)
    else:
        raise ValueError('Wrong number of parameters!')

    # no player specified
    if arg_count == 1 or arg_count == 3:
        # must be run by a player in this case because moving self
        if connection not in connection.protocol.players.values():
            raise ValueError("Both player and target player are required")
        player = connection.name
    # player specified
    elif arg_count == 2 or arg_count == 4:
        if not (connection.rights.admin or connection.rights.move_others):
            raise PermissionDenied(
                "moving other players requires the move_others right")
        player = args[0]

    player = get_player(connection.protocol, player)

    silent = connection.invisible or silent

    player.set_location((x, y, z))
    if connection is player:
        message = ('%s ' + ('silently ' if silent else '') + 'teleported to '
                   'location %s')
        message = message % (player.name, position)
    else:
        message = ('%s ' + ('silently ' if silent else '') + 'teleported %s '
                   'to location %s')
        message = message % (connection.name, player.name, position)
    if silent:
        connection.protocol.irc_say('* ' + message)
    else:
        connection.protocol.send_chat(message, irc=True)


@command(admin_only=True)
@target_player
def where(connection, player):
    """
    Tell you the coordinates of yourself or of a given player
    /where [player]
    """
    x, y, z = player.get_location()
    return '%s is in %s (%s, %s, %s)' % (
        player.name, to_coordinates(x, y), int(x), int(y), int(z))


@command('teleport', 'tp', admin_only=True)
def teleport(connection, player1, player2=None, silent=False):
    """
    Teleport yourself or a given player to another player
    /teleport [player] <target player>
    """
    # TODO: refactor this
    player1 = get_player(connection.protocol, player1)
    if player2 is not None:
        if connection.admin or connection.rights.teleport_other:
            player, target = player1, get_player(
                connection.protocol, player2)
            silent = silent or player.invisible
            message = ('%s ' + ('silently ' if silent else '') + 'teleported '
                       '%s to %s')
            message = message % (connection.name, player.name, target.name)
        else:
            return 'No administrator rights!'
    else:
        if connection not in connection.protocol.players.values():
            raise ValueError("Both player and target player are required")
        player, target = connection, player1
        silent = silent or player.invisible
        message = '%s ' + \
            ('silently ' if silent else '') + 'teleported to %s'
        message = message % (player.name, target.name)
    x, y, z = target.get_location()
    player.set_location(((x - 0.5), (y - 0.5), (z + 0.5)))
    if silent:
        connection.protocol.irc_say('* ' + message)
    else:
        connection.protocol.send_chat(message, irc=True)


@command('tpsilent', 'tps', admin_only=True)
def tpsilent(connection, player1, player2=None):
    """
    Silently teleport a player to another player
    /tpsilent [player] <target player>
    """
    teleport(connection, player1, player2, silent=True)


@command(admin_only=True)
@target_player
def fly(connection, player):
    """
    Enable flight
    /fly [player]
    Hold control and press space ;)
    """
    protocol = connection.protocol
    player.fly = not player.fly

    message = 'now flying' if player.fly else 'no longer flying'
    player.send_chat("You're %s" % message)
    if connection is not player and connection in protocol.players.values():
        connection.send_chat('%s is %s' % (player.name, message))
    protocol.irc_say('* %s is %s' % (player.name, message))



"""
Infiltration is an asymetric intel-based game mode where one team, the
attackers, tries to infiltrate the defenders base and steal the intel.
Defenders receive points for keeping the intel out of the attackers hands and
attackers recieve points for capturing the intel.

Options
^^^^^^^

.. code-block:: guess
    [infiltration]
    # Attackers get attacker_score_multiplier points for taking and capturing
    # the intel.
    attacker_score_multiplier = 10

    # Defenders gain 1 point for every defender_score_interval seconds that the
    # intel remains untouched.
    defender_score_interval = "30sec"

    # The ratio of attackers to defenders. Be aware that setting this
    # incorrectly might prevent players from joining
    attacker_ratio = 1.6

Originally created by: TheGrandmaster / hompy
"""

from twisted.internet.reactor import callLater
from twisted.internet.task import LoopingCall
from pyspades import contained as loaders
from pyspades.constants import *
from pyspades.player import create_player, player_left, intel_capture

from piqueserver.config import config, cast_duration

ATTACKER_TEAM = 1  # 0 = blue, 1 = green

INFIL_CONFIG = config.section("infiltration")
ATTACKER_RATIO_OPTION = INFIL_CONFIG.option(
    "attacker_ratio", default=1.6, cast=int)
ATTACKER_SCORE_MULTIPLIER_OPTION = INFIL_CONFIG.option(
    "attacker_score_multiplier", default=10, cast=int)
DEFENDER_SCORE_INTERVAL_OPTION = INFIL_CONFIG.option(
    "defender_score_interval", default="30sec", cast=cast_duration)

ATTACKER_SCORE_MULTIPLIER = 10
DEFENDER_SCORE_INTERVAL = 30  # seconds
ON_FLAG_TAKE_FLASHES_FOG = True

S_TEAM_FULL = 'Team full! The defending team always has less players'
S_ATTACKER_OBJECTIVE = '*** ATTACKERS: Infiltrate the enemy base and steal ' \
    'the intel!'
S_DEFENDER_OBJECTIVE = '*** DEFENDERS: Hold your ground! Earn points by ' \
    'keeping the intel safe'
S_OBJECTIVES = {
    ATTACKER_TEAM: S_ATTACKER_OBJECTIVE,
    1 - ATTACKER_TEAM: S_DEFENDER_OBJECTIVE
}


class DummyPlayer():
    def __init__(self, protocol, team):
        self.protocol = protocol
        self.team = team
        self.player_id = None
        self.acquire_player_id()

    def acquire_player_id(self):
        """temporarily borrow a player ID to score"""
        max_players = min(32, self.protocol.max_players)
        if len(self.protocol.connections) >= max_players:
            try:
                self.player_id = next(self.team.get_players()).player_id
            except StopIteration:
                self.player_id = None
            return self.player_id is not None
        self.player_id = self.protocol.player_ids.pop()
        self.protocol.player_ids.put_back(self.player_id)  # just borrowing it!
        create_player = loaders.CreatePlayer()
        create_player.x = 0
        create_player.y = 0
        create_player.z = 63
        create_player.weapon = RIFLE_WEAPON
        create_player.player_id = self.player_id
        create_player.name = self.team.name
        create_player.team = self.team.id
        self.protocol.broadcast_contained(create_player, save=True)
        return True

    def score(self):
        if self.protocol.game_mode != CTF_MODE:
            return
        if self.player_id in self.protocol.players:
            self.acquire_player_id()
        if self.player_id is None and not self.acquire_player_id():
            return
        winning = (self.protocol.max_score not in (0, None) and
                   self.team.score + 1 >= self.protocol.max_score)
        self.team.score += 1
        intel_capture = loaders.IntelCapture()
        intel_capture.player_id = self.player_id
        intel_capture.winning = winning
        self.protocol.broadcast_contained(intel_capture, save=True)
        if winning:
            self.team.initialize()
            self.team.other.initialize()
            for entity in self.protocol.entities:
                entity.update()
            for player in list(self.protocol.players.values()):
                player.hp = None
                if player.team is not None:
                    player.spawn()
            self.protocol.on_game_end()
        else:
            flag = self.team.other.set_flag()
            flag.update()

    def __del__(self):
        if self.player_id is None or self.player_id in self.protocol.players:
            return
        player_left = loaders.PlayerLeft()
        player_left.player_id = self.player_id
        self.protocol.broadcast_contained(player_left, save=True)


def apply_script(protocol, connection, config):
    class InfiltrationConnection(connection):

        def on_team_join(self, team):
            attacker = self.protocol.attacker
            defender = self.protocol.defender

            attacker_ratio = ATTACKER_RATIO_OPTION.get()
            attacker_count = attacker.count() + (1 if team is attacker else 0)
            defender_count = ((defender.count() +
                               (1 if team is defender else 0)) *
                              attacker_ratio)
            attacker_count = attacker_count or attacker_ratio
            if ((attacker_count > defender_count and team is attacker) or
                    (attacker_count < defender_count and team is defender)):
                if abs(attacker_count - defender_count) > attacker_ratio:
                    self.send_chat(S_TEAM_FULL)
                    return False
            return connection.on_team_join(self, team)

        def on_team_changed(self, old_team):
            if self.team and self.team.id in S_OBJECTIVES:
                self.send_chat(S_OBJECTIVES[self.team.id])
                self.send_chat_status(S_OBJECTIVES[self.team.id])
            connection.on_team_changed(self, old_team)

        def on_login(self, name):
            if self.team and self.team.id in S_OBJECTIVES:
                self.send_chat(S_OBJECTIVES[self.team.id])
                self.send_chat_status(S_OBJECTIVES[self.team.id])
            connection.on_login(self, name)

        def on_flag_capture(self):
            score_multiplier = ATTACKER_SCORE_MULTIPLIER_OPTION.get()
            if score_multiplier > 1:
                dummy = DummyPlayer(self.protocol, self.team)
                self.protocol.attacker_dummy = dummy
                if self.protocol.attacker_dummy_calls is None:
                    self.protocol.attacker_dummy_calls = []
                for i in range(score_multiplier - 1):
                    delay = i * 0.1
                    dummy_call = callLater(delay,
                                           self.protocol.attacker_dummy_score)
                    self.protocol.attacker_dummy_calls.append(dummy_call)
            self.protocol.cancel_defender_return_call()
            self.protocol.start_defender_score_loop()
            connection.on_flag_capture(self)

        def on_flag_take(self):
            if self.team is not self.protocol.attacker:
                return False
            if ON_FLAG_TAKE_FLASHES_FOG:
                self.protocol.fog_flash(self.team.color)
            if self.protocol.defender_score_loop.running:
                self.protocol.defender_score_loop.stop()
            return connection.on_flag_take(self)

        def on_flag_drop(self):
            self.protocol.start_defender_score_loop()
            connection.on_flag_drop(self)

    class InfiltrationProtocol(protocol):
        game_mode = CTF_MODE
        balanced_teams = None
        defender = None
        defender_score_loop = None
        defender_return_call = None
        attacker = None
        attacker_dummy = None
        attacker_dummy_calls = None

        def on_map_change(self, map):
            self.attacker = self.teams[ATTACKER_TEAM]
            self.defender = self.attacker.other
            self.defender_score_loop = LoopingCall(self.defender_score_cycle)
            self.start_defender_score_loop()
            protocol.on_map_change(self, map)

        def on_map_leave(self):
            if self.defender_score_loop and self.defender_score_loop.running:
                self.defender_score_loop.stop()
            self.defender_score_loop = None
            self.end_attacker_dummy_calls()
            protocol.on_map_leave(self)

        def on_game_end(self):
            self.end_attacker_dummy_calls()
            self.start_defender_score_loop()
            protocol.on_game_end(self)

        def on_flag_spawn(self, x, y, z, flag, entity_id):
            if flag.team is self.attacker:
                return 0, 0, 63
            return protocol.on_flag_spawn(self, x, y, z, flag, entity_id)

        def start_defender_score_loop(self):
            if self.defender_score_loop.running:
                self.defender_score_loop.stop()
            score_interval = DEFENDER_SCORE_INTERVAL_OPTION.get()
            self.defender_score_loop.start(score_interval, now=False)

        def defender_score_cycle(self):
            dummy = DummyPlayer(self, self.defender)
            dummy.score()

        def attacker_dummy_score(self):
            self.attacker_dummy.score()
            if self.attacker_dummy is None:
                # this could happen if the dummy capture caused the game to end
                return
            self.attacker_dummy_calls.pop(0)
            if not self.attacker_dummy_calls:
                self.end_attacker_dummy_calls()

        def end_attacker_dummy_calls(self):
            if self.attacker_dummy_calls:
                for dummy_call in self.attacker_dummy_calls:
                    if dummy_call and dummy_call.active():
                        dummy_call.cancel()
            self.attacker_dummy_calls = None
            self.attacker_dummy = None

        def cancel_defender_return_call(self):
            if self.defender_return_call and self.defender_return_call.active():
                self.defender_return_call.cancel()
            self.defender_return_call = None

        def fog_flash(self, color):
            old_color = self.get_fog_color()
            self.set_fog_color(color)
            callLater(0.2, self.set_fog_color, old_color)

    return InfiltrationProtocol, InfiltrationConnection



"""
Provides extensions to the map metadata (e.g. water damage).

.. codeauthor:: ?
"""


def apply_script(protocol, connection, config):
    class MapExtensionConnection(connection):

        def on_position_update(self):
            extensions = self.protocol.map_info.extensions
            if ('water_damage' in extensions and
                    self.world_object.position.z >= 61):
                water_damage = extensions['water_damage']
                self.environment_hit(water_damage)
            if 'boundary_damage' in extensions:
                x = self.world_object.position.x
                y = self.world_object.position.y
                boundary_damage = extensions['boundary_damage']
                if (x <= boundary_damage['left'] or x >= boundary_damage['right']
                        or y <= boundary_damage['top'] or y >= boundary_damage['bottom']):
                    self.environment_hit(boundary_damage['damage'])
            connection.on_position_update(self)

        def environment_hit(self, value):
            if value < 0 and self.hp >= 100:  # do nothing at max health
                return
            self.set_hp(self.hp - value)

        def on_command(self, command, parameters):
            disabled = self.protocol.map_info.extensions.get(
                'disabled_commands', [])
            if command in disabled:
                self.send_chat("Command '%s' disabled for this map" % command)
                return
            return connection.on_command(self, command, parameters)

    return protocol, MapExtensionConnection



# import everything here so at least we know commands will be registered
# also they are available to the rest of the code if required
from .game import get_time_limit, reset_game, lock, unlock, switch, set_balance, toggle_build, toggle_teamkill, global_chat, set_time_limit, fog
from .info import ping, rules, streak
from .map import mapname, show_rotation, change_planned_map, change_rotation, rotation_add, revert_rotation, advance
from .moderation import (get_ban_arguments, kick, ban, hban, dban, pban, banip,
                         unban, undo_ban, say, mute, unmute, ip, who_was,
                         invisible, godsilent, god, god_build)
from .movement import unstick, move_silent, move, where, teleport, tpsilent, fly
from .player import client, weapon, intel, kill, heal, deaf
from .server import server_name, server_info, version, scripts, toggle_master
from .social import login, pm, to_admin




import os
import importlib

from twisted.logger import Logger

log = Logger()


def check_scripts(script_names):
    '''Validation for a list of regular extension scripts.

    Check there are no duplicate scripts in the list of scripts names that is passed in.

    Args:
        script_names: The list of scripts names to be checked

    Return:
        bool: True if there are no duplicate scripts, False otherwise
    '''
    seen = set()
    dups = []
    for script in script_names:
        if script in seen:
            dups.append(script)
        else:
            seen.add(script)
    if dups:
        log.warn("Scripts included multiple times: {}".format(dups))
        return False
    return True


def check_game_mode(game_mode_name):
    '''Validation for a game mode script.

    Check if the game_mode is not a default one ('ctf' or 'tc').

    Args:
        game_mode_name: The game mode name to be checked

    Return:
        bool: True if the game mode is not a default one, False otherwise
    '''
    return game_mode_name not in ('ctf', 'tc')


def load_scripts(script_names, script_dir, script_type):
    '''Load script as module.

    Loads all scripts, matching the script_names, from the script_dir folder. The function also requires the
    specification of which type of scripts are passed in (e.g. "script", "gamemode", "testscript"). This is
    necessary for naming the namespace and error handling.

    Args:
        script_names: The list of script names to be loaded
        script_dir: The path to the corresponding scripts directory
        script_type: The script type ("script" for regular scripts and "gamemode" for game_mode script)

    Return:
        [module]: The list of module objects containing the loaded scripts
    '''
    script_objects = []

    finder = importlib.machinery.PathFinder()
    for script in script_names:
        spec_scripts = finder.find_spec(script, [script_dir])
        spec_global = importlib.util.find_spec(script)
        spec = spec_scripts or spec_global
        if not spec:
            log.error(
                "{} '{}' not found in either {} directory or global scope".format(
                    script_type, script, script_dir))
            continue
        # namespace module name to avoid shadowing global modules
        # TODO: figure out if there are any right or better ways.
        spec.name = 'piqueserver_{}_namespace_{}'.format(script_type, script)
        spec.loader.name = spec.name
        # load module
        try:
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            script_objects.append(module)
            continue
        except Exception as e: # needs to be broad since we exec the module
            log.error("Error while loading {} {}: {!r}".format(
                script_type, script, e))

    return script_objects


def load_scripts_regular_extension(script_names, script_dir):
    ''' Wrapper for load function

    It loads scripts in the case of regular extension scripts.

    Args:
        script_names: The list of script names to be loaded
        script_dir: The path to the corresponding scripts directory

    Return:
        [module]: The list of module objects containing the loaded scripts

    '''
    return load_scripts(script_names, script_dir, 'script')


def load_script_game_mode(script_name, script_dir):
    ''' Wrapper for load function

    It loads scripts in the case of game mode scripts. Prior to this, it checks if the game mode is not a
    default one (if it's a default no scripts should be loaded)

    Args:
        script_names: The list of script names to be loaded
        script_dir: The path to the corresponding scripts directory

    Return:
        [module]: The list of module objects containing the uploaded scripts

    '''
    if check_game_mode(script_name):
        return load_scripts([script_name], script_dir, 'gamemode')
    return []


def apply_scripts(scripts, config, protocol_class, connection_class):
    ''' Application of scripts modules

    It applies the script modules to the specified protocol and connection class instances, in order to build
    the resulting classes with all the additional features that the scripts implement (more information
    about how scripts are implemented and used can be found in the project documentation
    https://piqueserver.readthedocs.io/en/latest/architecture.html#extension-scripts).

    Args:
        scripts: List of scripts modules to apply
        config: It holds the dict of sections and options; this is required by the scripts
         logic
        protocol_class: The protocol class instance to update
        connection_class: The connection class instance to update

    Return:
        (FeatureProtocol, FeatureConnection): The updated protocol and connection class instances
    '''

    for script in scripts:
        protocol_class, connection_class = script.apply_script(
            protocol_class, connection_class, config.get_dict())

    return (protocol_class, connection_class)



from pyspades.constants import SPADE_TOOL
from piqueserver.commands import command


@command('togglewallspading', 'tws', admin_only=True)
def togglewallspading(self):
    self.protocol.is_fluffy = not self.protocol.is_fluffy
    if not self.protocol.is_fluffy:
        return 'Spading someone behind a wall is possible again now!'
    else:
        return 'You canot spade people through walls anymore. (killjoy)'


def apply_script(protocol, connection, config):

    class nowallspadingprotocol(protocol):
        is_fluffy = True

    class nowallspadingconnection(connection):

        def on_hit(self, hit_amount, hit_player, hit_type, grenade):

            if hit_player.protocol.is_fluffy:
                pos = hit_player.world_object.position

                if not self.world_object.can_see(
                        pos.x,
                        pos.y,
                        pos.z) and self.tool == SPADE_TOOL and not grenade and self.team != hit_player.team:
                    return False

            return connection.on_hit(
                self, hit_amount, hit_player, hit_type, None)

    return nowallspadingprotocol, nowallspadingconnection



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

cdef class Loader:
    def __init__(self, ByteReader reader = None):
        if reader is not None:
            self.read(reader)

    cpdef read(self, ByteReader reader):
        read_python = getattr(self, 'read', None)
        if read_python is None:
            raise NotImplementedError('read() not implemented')
        read_python(reader)

    cpdef write(self, ByteWriter writer):
        write_python = getattr(self, 'write', None)
        if write_python is None:
            raise NotImplementedError('write() not implemented')
        write_python(writer)

    cpdef ByteWriter generate(self):
        cdef ByteWriter writer = ByteWriter()
        self.write(writer)
        return writer



# Copyright (c) James W. Hofmann 2011.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import json

from twisted.internet import reactor
from twisted.web import server
from twisted.web.resource import Resource


class PublishResource(Resource):

    def __init__(self, factory):
        self.factory = factory
        Resource.__init__(self)

    def getChild(self, name, request):
        return self

    def render_GET(self, request):
        return self.factory.json_bans


class PublishServer:

    def __init__(self, protocol, port):
        self.protocol = protocol
        publish_resource = PublishResource(self)
        site = server.Site(publish_resource)
        protocol.listenTCP(port, site)
        self.update()

    def update(self):
        bans = []
        for network, (_name, reason, timestamp) in self.protocol.bans.iteritems():
            if timestamp is None or reactor.seconds() < timestamp:
                bans.append({"ip": network, "reason": reason})
        self.json_bans = json.dumps(bans)



"""
Helps Deuces automagically when they ask in the chat for help.

.. codeauthor:: ?
"""

import re
from twisted.internet import reactor

deuce_name_pattern = re.compile(r"Deuce\d?\d?$")
nick_chat_pattern = re.compile(
    ".*how.*(set|change|choo?se|make|pick).*(name|nick)",
    re.IGNORECASE)
airstrike_chat_pattern = re.compile(
    ".*how.*(to|make|use|get).*(airstrike|killstreak|airsupport)",
    re.IGNORECASE)


def deuce_howto_match(player, msg):
    return (not deuce_name_pattern.match(player.name) is None and
            not nick_chat_pattern.match(msg) is None)


def airstrike_howto_match(player, msg):
    return (not airstrike_chat_pattern.match(msg) is None)


def apply_script(protocol, connection, config):
    def send_help_nick(connection):
        connection.protocol.send_chat(
            "TO CHANGE YOUR NAME: Start Menu-> "
            "All Programs-> Ace of Spades-> Configuration")
        connection.protocol.irc_say("* Sent nick help to %s" % connection.name)

    def send_help_airstrike(connection):
        connection.protocol.send_chat(
            "TO USE AN AIRSTRIKE: Once you have 15 points, "
            "get a 6 killstreak ->                  "
            "Then type /airstrike G4 if you want the strike to hit G4")
        connection.protocol.irc_say(
            "* Sent airstrike help to %s" % connection.name)

    class AutoHelpConnection(connection):

        def on_chat(self, value, global_message):
            if deuce_howto_match(self, value):
                reactor.callLater(1.0, send_help_nick, self)
            if airstrike_howto_match(self, value):
                reactor.callLater(1.0, send_help_airstrike, self)
            return connection.on_chat(self, value, global_message)
    return protocol, AutoHelpConnection



"""
Allows players to vote for maps.

Commands
^^^^^^^^

* ``/votemap`` initiates map voting
* ``/vote <map name>`` vote for a map

.. code-block:: guess

    [votemap]
    public_votes = true
    extension_time = "15min"
    player_driven = false
    autoschedule = false
    percentage = 80

.. codeauthor:: James Hofmann a.k.a triplefox (GPL LICENSE)
"""


import random
from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from pyspades.common import prettify_timespan
from piqueserver.map import check_rotation
from piqueserver.scheduler import Scheduler
from piqueserver.commands import command, player_only
from piqueserver.config import config, cast_duration

votemap_config = config.section('votemap')

VOTEMAP_AUTOSCHEDULE_OPTION = votemap_config.option('autoschedule', 180)
VOTEMAP_PUBLIC_VOTES_OPTION = votemap_config.option('public_votes', True)
# godwhoa: This option gets loaded into votemap_time but that doesn't get used anywhere.
VOTEMAP_TIME_OPTION = votemap_config.option('time', default="2min", cast=cast_duration)
VOTEMAP_EXTENSION_TIME_OPTION = votemap_config.option('extension_time', default="15min",
    cast=lambda x: cast_duration(x)/60)
VOTEMAP_PLAYER_DRIVEN_OPTION = votemap_config.option('player_driven', False)
VOTEMAP_PERCENTAGE_OPTION = votemap_config.option('percentage', 80)

def cancel_verify(connection, instigator):
    return (connection.admin or
            connection is instigator or
            connection.rights.cancel)


class VoteMap:
    extension_time = 0.0
    public_votes = True
    instigator = None
    protocol = None

    def __init__(self, connection, protocol, rotation):
        self.instigator = connection
        self.rotation = rotation
        self.protocol = protocol
        self.vote_percentage = self.protocol.votemap_percentage
        self.vote_time = self.protocol.votemap_time
        self.vote_interval = self.protocol.votemap_interval
        self.public_votes = self.protocol.votemap_public_votes
        self.extension_time = self.protocol.votemap_extension_time
        rotation = [n.name.lower() for n in rotation]
        final_rotation = random.sample(rotation, min(len(rotation), 5))
        if self.extension_time > 0:
            final_rotation.append("extend")
        self.picks = final_rotation
        self.votes = {}
        if connection is None:
            name = 'server'
        else:
            name = connection.name

    def votes_left(self):
        thresh = int((len(self.protocol.players)) *
                     self.vote_percentage / 100.0)
        counts = {}
        for v in list(self.votes.values()):
            if v in counts:
                counts[v]['count'] += 1
            else:
                counts[v] = {'name': v, 'count': 1}
        cvlist = list(counts.values())
        if len(cvlist) <= 0:
            return {'name': self.picks[0], 'count': 0}
        mv = cvlist[0]
        for n in list(counts.keys()):
            if counts[n]['count'] > mv['count']:
                mv = n
        mv['count'] = thresh - mv['count']
        return mv

    def verify(self):
        instigator = self.instigator
        if instigator is None:
            return True
        last = instigator.last_votemap
        if (last is not None and
                reactor.seconds() - last < self.vote_interval):
            return "You can't start a vote now."
        return True

    def start(self):
        instigator = self.instigator
        protocol = self.protocol
        if instigator is None:
            protocol.send_chat('Time to vote!', irc=True)
        else:
            protocol.send_chat(
                '* %s initiated a map vote.' % instigator.name, irc=True)
        self.schedule = schedule = Scheduler(protocol)
        schedule.call_later(self.vote_time, self.timeout)
        schedule.loop_call(30.0, self.update)
        self.protocol.votemap = self
        self.update()

    def vote(self, connection, mapname):
        mapname = mapname.lower()
        if mapname not in self.picks:
            connection.send_chat("Map %s is not available." % mapname)
            return
        self.votes[connection] = mapname
        if self.public_votes:
            self.protocol.send_chat('%s voted for %s.' % (connection.name,
                                                          mapname))
        if self.votes_left()['count'] <= 0:
            self.on_majority()

    def cancel(self, connection=None):
        if connection is None:
            message = 'Cancelled'
        elif not cancel_verify(connection, self.instigator):
            return 'You did not start the vote.'
        else:
            message = 'Cancelled by %s' % connection.name
        self.protocol.send_chat(message)
        self.set_cooldown()
        self.finish()

    def update(self):
        self.protocol.send_chat(
            'Choose next map. Say /vote <name> to cast vote.')
        names = ' '.join(self.picks)
        self.protocol.send_chat('Maps: %s' % names)
        self.protocol.send_chat('To extend current map: /vote extend')

    def timeout(self):
        self.show_result()
        self.finish()

    def on_majority(self):
        self.show_result()
        self.finish()

    def show_result(self):
        result = self.votes_left()['name']
        if result == "extend":
            tl = self.protocol.set_time_limit(self.extension_time, True)
            span = prettify_timespan(tl * 60.0)
            self.protocol.send_chat('Mapvote ended. Current map will '
                                    'continue for %s.' % span, irc=True)
            self.protocol.autoschedule_votemap()
        else:
            self.protocol.send_chat('Mapvote ended. Next map will be: %s.' %
                                    result, irc=True)
            self.protocol.planned_map = check_rotation([result])[0]
        self.set_cooldown()

    def set_cooldown(self):
        if self.instigator is not None and not self.instigator.admin:
            self.instigator.last_votemap = reactor.seconds()

    def finish(self):
        self.schedule.reset()
        self.protocol.votemap = None


@command()
@player_only
def votemap(connection, *arg):
    if not connection.protocol.votemap_player_driven and not connection.admin:
        return "Player-initiated mapvotes are disabled on this server."
    return connection.protocol.start_votemap(
        VoteMap(connection, connection.protocol, connection.protocol.maps))


@command('vote')
@player_only
def votemap_vote(connection, value):
    if connection.protocol.votemap is not None:
        return connection.protocol.votemap.vote(connection, value)
    else:
        return 'No map vote in progress.'


def apply_script(protocol, connection, config):
    class VoteProtocol(protocol):
        # voting
        votemap_time = 120
        votemap_interval = 3 * 60
        votemap_percentage = 80.0
        votemap = None
        planned_map = None
        autoschedule_call = None

        # voting

        def __init__(self, interface, config):
            protocol.__init__(self, interface, config)
            self.votemap_autoschedule = VOTEMAP_AUTOSCHEDULE_OPTION.get()
            self.votemap_public_votes = VOTEMAP_PUBLIC_VOTES_OPTION.get()
            self.votemap_time = VOTEMAP_TIME_OPTION.get()
            self.votemap_extension_time = VOTEMAP_EXTENSION_TIME_OPTION.get()
            self.votemap_player_driven = VOTEMAP_PLAYER_DRIVEN_OPTION.get()
            self.votemap_percentage = VOTEMAP_PERCENTAGE_OPTION.get()
            self.autoschedule_votemap()

        def autoschedule_votemap(self):
            if self.votemap_autoschedule > 0 and self.autoschedule_call is None:
                self.autoschedule_call = self.call_end(
                    self.votemap_autoschedule, self.start_votemap)

        def cancel_vote(self, connection=None):
            if self.votemap is not None:
                return self.votemap.cancel(connection)
            else:
                return protocol.cancel_vote(self, connection)

        def start_votemap(self, votemap=None):
            if self.votemap is not None:
                return self.votemap.update()
            if votemap is None:
                votemap = VoteMap(None, self, self.maps)
            verify = votemap.verify()
            if verify is True:
                votemap.start()
            else:
                return verify

        def on_advance(self, *arg, **kw):
            protocol.on_advance(self, *arg, **kw)
            self.end_votes()

        def end_votes(self):
            if self.votemap is not None:
                self.votemap.finish()

    class VoteConnection(connection):
        last_votemap = None

    return VoteProtocol, VoteConnection



"""
Protects areas against block destroying/building.

Commands
^^^^^^^^

* ``/protect <area coordinates>`` puts an area under protected status *admin only*
* ``/protect`` clears all protected areas *admin only*

.. codeauthor:: hompy
"""

from piqueserver.commands import command, admin
from pyspades.common import coordinates


@command(admin_only=True)
def protect(connection, value=None):
    protocol = connection.protocol
    if value is None:
        protocol.protected = None
        protocol.send_chat('All areas unprotected', irc=True)
    else:
        if protocol.protected is None:
            protocol.protected = set()
        pos = coordinates(value)
        protocol.protected.symmetric_difference_update([pos])
        message = 'The area at %s is now %s' % (
            value.upper(), 'protected' if pos in protocol.protected else 'unprotected')
        protocol.send_chat(message, irc=True)


def apply_script(protocol, connection, config):
    class ProtectConnection(connection):

        def _block_available(self, x, y, z):
            if not self.god and self.protocol.is_protected(x, y, z):
                return False

        def on_block_build_attempt(self, x, y, z):
            if not self.god and self.protocol.is_protected(x, y, z):
                return False
            return connection.on_block_build_attempt(self, x, y, z)

        def on_line_build_attempt(self, points):
            if not self.god:
                for point in points:
                    if self.protocol.is_protected(*point):
                        return False
            return connection.on_line_build_attempt(self, points)

    class ProtectProtocol(protocol):
        protected = None

        def on_map_change(self, map):
            self.protected = set(coordinates(s) for s in
                                 getattr(self.map_info.info, 'protected', []))
            protocol.on_map_change(self, map)

        def is_indestructable(self, x, y, z):
            if self.is_protected(x, y, z):
                return True
            return protocol.is_indestructable(self, x, y, z)

        def is_protected(self, x, y, z):
            if self.protected:
                for sx, sy in self.protected:
                    if x >= sx and y >= sy and x < sx + 64 and y < sy + 64:
                        return True
            return False

    return ProtectProtocol, ProtectConnection



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import random
from itertools import product
import enet

from pyspades.protocol import BaseProtocol
from pyspades.constants import (
    CTF_MODE, TC_MODE, GAME_VERSION, MIN_TERRITORY_COUNT, MAX_TERRITORY_COUNT,
    UPDATE_FREQUENCY, UPDATE_FPS, NETWORK_FPS)
from pyspades.types import IDPool
from pyspades.master import get_master_connection
from pyspades.team import Team
from pyspades.entities import Territory
# importing tc_data is a quick hack since this file writes into it
from pyspades.player import ServerConnection, tc_data
from pyspades import world
from pyspades.bytes import ByteWriter
from pyspades import contained as loaders
from pyspades.common import make_color
from pyspades.mapgenerator import ProgressiveMapGenerator



class ServerProtocol(BaseProtocol):
    connection_class = ServerConnection

    name = 'pyspades server'
    game_mode = CTF_MODE
    max_players = 32
    connections = None
    player_ids = None
    master = False
    max_score = 10
    map = None
    spade_teamkills_on_grief = False
    friendly_fire = False
    friendly_fire_time = 2
    server_prefix = '[*] '
    respawn_time = 5
    refill_interval = 20
    master_connection = None
    speedhack_detect = True
    fog_color = (128, 232, 255)
    winning_player = None
    world = None
    team_class = Team
    team1_color = (0, 0, 196)
    team2_color = (0, 196, 0)
    team1_name = 'Blue'
    team2_name = 'Green'
    spectator_name = 'Spectator'
    loop_count = 0
    melee_damage = 100
    version = GAME_VERSION
    respawn_waves = False

    def __init__(self, *arg, **kw):
        # +2 to allow server->master and master->server connection since enet
        # allocates peers for both clients and hosts. this is done at
        # enet-level, not application-level, so even for masterless-servers,
        # this should not allow additional players.
        self.max_connections = self.max_players + 2
        BaseProtocol.__init__(self, *arg, **kw)
        self.entities = []
        self.players = {}
        self.player_ids = IDPool()

        self._create_teams()

        self.world = world.World()
        self.set_master()

        # safe position LUT
        #
        # Generates a LUT to check for safe positions. The slighly weird
        # sorting is used to sort by increasing distance so the nearest spots
        # get chosen first
        # product(repeat=3) is the equivalent of 3 nested for loops
        self.pos_table = list(product(range(-5, 6), repeat=3))

        self.pos_table.sort(key=lambda vec: abs(vec[0] * 1.03) +
                            abs(vec[1] * 1.02) +
                            abs(vec[2] * 1.01))

    def _create_teams(self):
        """create the teams
        This Method is separate to simplify unit testing
        """
        self.team_spectator = self.team_class(-1, self.spectator_name,
                                              (0, 0, 0), True, self)
        self.team_1 = self.team_class(0, self.team1_name, self.team1_color,
                                      False, self)
        self.team_2 = self.team_class(1, self.team2_name, self.team2_color,
                                      False, self)
        self.teams = {
            -1: self.team_spectator,
            0: self.team_1,
            1: self.team_2
        }
        self.team_1.other = self.team_2
        self.team_2.other = self.team_1

    @property
    def blue_team(self):
        """alias to team_1 for backwards-compatibility"""
        return self.team_1

    @property
    def green_team(self):
        """alias to team_2 for backwards-compatibility"""
        return self.team_2

    @property
    def spectator_team(self):
        """alias to team_spectator for backwards-compatibility"""
        return self.team_spectator

    def broadcast_contained(self, contained, unsequenced=False, sender=None,
                            team=None, save=False, rule=None):
        """send a Contained `Loader` to all or a selection of connected
        players

        Parameters:
            contained: the `Loader` object to send
            unsequenced: set the enet ``UNSEQUENCED`` flag on this packet
            sender: if set to a connection object, do not send this packet to
                that player, as they are the sender.
            team: if set to a team, only send the packet to that team
            save: if the player has not downloaded the map yet, save this
                packet and send it when the map transfer has completed
            rule: if set to a callable, this function is called with the player
                as parameter to determine if a given player should receive the
                packet
        """
        if unsequenced:
            flags = enet.PACKET_FLAG_UNSEQUENCED
        else:
            flags = enet.PACKET_FLAG_RELIABLE
        writer = ByteWriter()
        contained.write(writer)
        data = bytes(writer)
        packet = enet.Packet(data, flags)
        for player in self.connections.values():
            if player is sender or player.player_id is None:
                continue
            if team is not None and player.team is not team:
                continue
            if rule is not None and not rule(player):
                continue
            if player.saved_loaders is not None:
                if save:
                    player.saved_loaders.append(data)
            else:
                player.peer.send(0, packet)

    # backwards compatability
    send_contained = broadcast_contained

    def reset_tc(self):
        self.entities = self.get_cp_entities()
        for entity in self.entities:
            team = entity.team
            if team is None:
                entity.progress = 0.5
            else:
                team.score += 1
                entity.progress = float(team.id)
        tc_data.set_entities(self.entities)
        self.max_score = len(self.entities)

    def get_cp_entities(self):
        # cool algorithm number 1
        entities = []
        land_count = self.map.count_land(0, 0, 512, 512)
        territory_count = int((land_count / (512.0 * 512.0)) * (
            MAX_TERRITORY_COUNT - MIN_TERRITORY_COUNT) + MIN_TERRITORY_COUNT)
        j = 512.0 / territory_count
        for i in range(territory_count):
            x1 = i * j
            y1 = 512 / 4
            x2 = (i + 1) * j
            y2 = y1 * 3
            flag = Territory(i, self, *self.get_random_location(
                zone=(x1, y1, x2, y2)))
            if i < territory_count / 2:
                team = self.team_1
            elif i > (territory_count - 1) / 2:
                team = self.team_2
            else:
                # odd number - neutral
                team = None
            flag.team = team
            entities.append(flag)
        return entities

    def update(self):
        self.loop_count += 1
        BaseProtocol.update(self)
        for player in self.connections.values():
            if (player.map_data is not None and
                    not player.peer.reliableDataInTransit):
                player.continue_map_transfer()
        self.world.update(UPDATE_FREQUENCY)
        self.on_world_update()
        if self.loop_count % int(UPDATE_FPS / NETWORK_FPS) == 0:
            self.update_network()

    def update_network(self):
        items = []
        highest_player_id = 0
        for i in range(32):
            position = orientation = None
            try:
                player = self.players[i]
                highest_player_id = i
                if (not player.filter_visibility_data and
                        not player.team.spectator):
                    world_object = player.world_object
                    position = world_object.position.get()
                    orientation = world_object.orientation.get()
            except (KeyError, TypeError, AttributeError):
                pass
            if position is None:
                position = (0.0, 0.0, 0.0)
                orientation = (0.0, 0.0, 0.0)
            items.append((position, orientation))
        world_update = loaders.WorldUpdate()
        # we only want to send as many items of the player list as needed, so
        # we slice it off at the highest player id
        world_update.items = items[:highest_player_id+1]
        self.send_contained(world_update, unsequenced=True)

    def set_map(self, map_obj):
        self.map = map_obj
        self.world.map = map_obj
        self.on_map_change(map_obj)
        self.team_1.initialize()
        self.team_2.initialize()
        if self.game_mode == TC_MODE:
            self.reset_tc()
        self.players = {}
        if self.connections:
            data = ProgressiveMapGenerator(self.map, parent=True)
            for connection in list(self.connections.values()):
                if connection.player_id is None:
                    continue
                if connection.map_data is not None:
                    connection.disconnect()
                    continue
                connection.reset()
                connection._send_connection_data()
                connection.send_map(data.get_child())
        self.update_entities()

    def reset_game(self, player=None, territory=None):
        """reset the score of the game

        player is the player which should be awarded the necessary captures to
        end the game
        """
        self.team_1.initialize()
        self.team_2.initialize()
        if self.game_mode == CTF_MODE:
            if player is None:
                player = list(self.players.values())[0]
            intel_capture = loaders.IntelCapture()
            intel_capture.player_id = player.player_id
            intel_capture.winning = True
            self.send_contained(intel_capture, save=True)
        elif self.game_mode == TC_MODE:
            if territory is None:
                territory = self.entities[0]
            territory_capture = loaders.TerritoryCapture()
            territory_capture.object_index = territory.id
            territory_capture.winning = True
            territory_capture.state = territory.team.id
            self.send_contained(territory_capture)
            self.reset_tc()
        for entity in self.entities:
            entity.update()
        for player in self.players.values():
            if player.team is not None:
                player.spawn()

    def get_name(self, name):
        '''
        Sanitizes `name` and modifies it so that it doesn't collide with other names connected to the server.

        Returns the fixed name.
        '''
        name = name.replace('%', '')
        new_name = name
        names = [p.name.lower() for p in self.players.values()]
        i = 0
        while new_name.lower() in names:
            i += 1
            new_name = name + str(i)
        return new_name

    def get_mode_mode(self):
        if self.game_mode == CTF_MODE:
            return 'ctf'
        elif self.game_mode == TC_MODE:
            return 'tc'
        return 'unknown'

    def get_random_location(self, force_land=True, zone=(0, 0, 512, 512)):
        x1, y1, x2, y2 = zone
        if force_land:
            x, y = self.map.get_random_point(x1, y1, x2, y2)
        else:
            x = random.randrange(x1, x2)
            y = random.randrange(y1, y2)
        z = self.map.get_z(x, y)
        return x, y, z

    def set_master(self):
        if self.master:
            get_master_connection(self).addCallbacks(
                self.got_master_connection,
                self.master_disconnected)

    def got_master_connection(self, connection):
        self.master_connection = connection
        connection.disconnect_callback = self.master_disconnected
        self.update_master()

    def master_disconnected(self, client=None):
        self.master_connection = None

    def get_player_count(self):
        count = 0
        for connection in self.connections.values():
            if connection.player_id is not None:
                count += 1
        return count

    def update_master(self):
        if self.master_connection is None:
            return
        self.master_connection.set_count(self.get_player_count())

    def update_entities(self):
        map_obj = self.map
        for entity in self.entities:
            moved = False
            if map_obj.get_solid(entity.x, entity.y, entity.z - 1):
                moved = True
                entity.z -= 1
                # while solid in block above (ie. in the space in which the
                # entity is sitting), move entity up)
                while map_obj.get_solid(entity.x, entity.y, entity.z - 1):
                    entity.z -= 1
            else:
                # get_solid can return None, so a more specific check is used
                while map_obj.get_solid(entity.x, entity.y, entity.z) is False:
                    moved = True
                    entity.z += 1
            if moved or self.on_update_entity(entity):
                entity.update()

    def broadcast_chat(self, message, global_message=None, sender=None,
                       team=None):
        for player in self.players.values():
            if player is sender:
                continue
            if player.deaf:
                continue
            if team is not None and player.team is not team:
                continue
            player.send_chat(message, global_message)

    # backwards compatability
    send_chat = broadcast_chat

    def broadcast_chat_warning(self, message, team=None):
        """
        Send a warning message. This gets displayed
        as a yellow popup with sound for OpenSpades
        clients
        """
        self.send_chat(self, "%% " + str(message), team=team)

    def broadcast_chat_notice(self, message, team=None):
        """
        Send a warning message. This gets displayed
        as a popup for OpenSpades
        clients
        """
        self.send_chat(self, "N% " + str(message), team=team)

    def broadcast_chat_error(self, message, team=None):
        """
        Send a warning message. This gets displayed
        as a red popup with sound for OpenSpades
        clients
        """
        self.send_chat(self, "!% " + str(message), team=team)

    def broadcast_chat_status(self, message, team=None):
        """
        Send a warning message. This gets displayed as a message in the status
        area at the top of the screen, where events such as intel pickups are
        also displayed.
        """
        self.send_chat(self, "C% " + str(message), team=team)

    def set_fog_color(self, color):
        self.fog_color = color
        fog_color = loaders.FogColor()
        fog_color.color = make_color(*color)
        self.send_contained(fog_color, save=True)

    def get_fog_color(self):
        return self.fog_color

    # events

    def on_cp_capture(self, cp):
        pass

    def on_game_end(self):
        pass

    def on_world_update(self):
        pass

    def on_map_change(self, map_):
        pass

    def on_base_spawn(self, x, y, z, base, entity_id):
        pass

    def on_flag_spawn(self, x, y, z, flag, entity_id):
        pass

    def on_update_entity(self, entity):
        pass



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=too-many-lines
# pylint: disable=wildcard-import,unused-wildcard-import

import traceback
import inspect
import warnings
from collections import namedtuple
import textwrap
import functools
from typing import Dict, List, Callable

from twisted.logger import Logger

from pyspades.common import escape_control_codes
from pyspades.player import parse_command

_commands = {}
_alias_map = {}
_rights = {}  # type: Dict[str, List[str]]

log = Logger()

class CommandError(Exception):
    pass


class PermissionDenied(Exception):
    pass


def command(name=None, *aliases,
            admin_only=False) -> Callable[[Callable], Callable]:
    """
    Register a new command.

    The command will be accessible as `/function_name` unless a name or alias
    is specified.

    >>> @command()
    ... def some_command(x):
    ...     pass

    Optional names and aliases:

    >>> @command("name", "alias1", "alias2")
    ... def some_command(x):
    ...     pass
    """
    def decorator(function) -> Callable[..., str]:
        function.user_types = set()
        if admin_only:
            function.user_types.add("admin")

        nonlocal name
        if name is None:
            name = function.__name__

        function.command_name = name

        _commands[name] = function

        if aliases:
            function.aliases = []

            for alias in aliases:
                _alias_map[alias] = name

        return function
    return decorator


def add(func: Callable) -> None:
    """
    Function to add a command from scripts. Deprecated
    """
    warnings.warn(
        '@add is deprecated, use @command()',
        DeprecationWarning)
    command()(func)


def name(name: str) -> Callable:
    """
    Give the command a new name. Deprecated
    """
    warnings.warn(
        '@name is deprecated, use @command("name")',
        DeprecationWarning)

    def dec(func: Callable) -> Callable:
        func.__name__ = name
        return func
    return dec


def alias(name: str) -> Callable:
    """
    add a new alias to a command. Deprecated
    """
    warnings.warn(
        '@alias is deprecated, use @command("name", "alias1", "alias2")',
        DeprecationWarning)

    def dec(func: Callable) -> Callable:
        try:
            func.aliases.append(name)
        except AttributeError:
            func.aliases = [name]
        return func
    return dec


def restrict(*user_types: List[str]) -> Callable:
    """
    restrict the command to only be used by a specific type of user

    >>> @restrict("admin", "guard")
    ... @command()
    ... def some_command(x):
    ...     pass
    """
    def decorator(function: Callable) -> Callable:
        function.user_types = set(*user_types)
        return function
    return decorator


def has_permission(f, connection):
    if not f.user_types:
        return True
    elif f.command_name in connection.rights:
        return True
    elif connection.admin:
        return True
    else:
        return False


CommandHelp = namedtuple("CommandHelp", ["description", "usage", "info"])


def get_command_help(command_func: Callable) -> CommandHelp:
    doc = command_func.__doc__
    if not doc:
        return CommandHelp("", "", "")

    doc = textwrap.dedent(doc).strip()

    lines = doc.split("\n")

    # Yes, this is disgusting. I wish I could think of a better way of
    # doing it

    count = 0
    desc = ""
    usage = ""
    info = ""

    for line in lines:
        line = line.strip()
        if not line:
            continue
        if count == 0:
            desc = line
        elif count == 1:
            usage = line
        elif count == 2:
            info = line

        count += 1

    return CommandHelp(desc, usage, info)


def format_command_error(command_func: Callable, message: str, exception:
                         Exception=None) -> str:
    """format a help message for a given command"""
    command_help = get_command_help(command_func)

    if command_help.usage:
        return "{}\nUsage: {}".format(message, command_help.usage)
    else:
        return message

# it might seem unecesary to reimplement half of the dict data type for
# _rights, but previously this attribute was accessed and modified from
# various places and that made things more confusing than the needed to be


def add_rights(user_type: str, command_name: str) -> None:
    """
    Give the user type a new right

    >>> add_rights("knights", "say_ni")
    """
    _rights.setdefault(user_type, []).append(command_name)


def get_rights(user_type: str) -> List[str]:
    """
    Get a list of rights a specific user type has.

    >>> add_rights("knights", "say_ni")
    >>> get_rights("knights")
    ["say_ni"]
    >>> get_rights("arthur")
    []
    """
    r = _rights.get(user_type, [])
    return r


def update_rights(rights: Dict):
    """
    Update the rights of all users according to the input dictionary. This
    is currently only here for when the config needs to be reloaded.

    >>> update_rights({"knights": ["say_ni"]})
    >>> get_rights("knights")
    ["say_ni"]
    """
    _rights.update(rights)


def admin(func: Callable) -> Callable:
    """
    Shorthand for @restrict("admin"). Mainly exists for backwards
    compatibility with pyspades scripts.

    >>> @admin
    ... @command()
    ... def some_command(x):
    ...     pass
    """
    return restrict('admin')(func)


def player_only(func: Callable):
    """This recorator restricts a command to only be runnable by players
    connected to the server, not via other places such as, say, the console.

    >>> @command()
    ... @player_only
    ... def some_command(x):
    ...     pass
    """
    @functools.wraps(func)
    def _decorated(connection, *args, **kwargs):
        if connection not in connection.protocol.players.values():
            raise CommandError("only players can't use this command")
        func(connection, *args, **kwargs)
    return _decorated

def target_player(func: Callable):
    """This decorator converts first argument of a command to a `piqueserver.FeatureConnection`.
       It's intended for commands which accept single argument for target player eg. /fly [player].
       It implicitly uses invoker as target if no arguments are provided.
       It uses first argument are player name or id for targetting.
       It forces non-player invokers to provide player argument.

    >>> @command()
    ... @target_player
    ... def fly(connection, target):
    ...     target.fly = True
    ...     pass
    """
    @functools.wraps(func)
    def _decorated(connection, *args, **kwargs):
        is_player = connection in connection.protocol.players.values()
        # implicitly set target to invoker if no args
        if len(args) == 0 and is_player:
            args = (connection,)
        # try and use first arg as player name or id to target
        elif len(args) > 0:
            args = (get_player(connection.protocol, args[0]), *args[1:])
        # console or irc invokers are required to provide a target
        else:
            raise ValueError("Target player is required")
        return func(connection, *args, **kwargs)
    return _decorated

def get_player(protocol, value: str, spectators=True):
    """
    Get a player connection object by name or ID.

    IDs are formatted as: "#12"

    If no player with the specified name is found, it will try to find a
    player who's name contains the string
    """
    player_obj = None
    try:
        if value.startswith('#'):
            value = int(value[1:])
            player_obj = protocol.players[value]
        else:
            players = protocol.players
            try:
                player_obj = players[value]
            except KeyError:
                value = value.lower()
                matches = []
                for player in players.values():
                    name = player.name.lower()
                    if name == value:
                        # if the playername matches, we want to return directly, otherwise
                        # you would not be able to kick players who's name is a subset of
                        # another players name
                        return player
                    if name.count(value):
                        matches.append(player)

                if len(matches) > 1:
                    # more than one player matched
                    raise CommandError("Ambiguous player")
                elif len(matches) == 1:
                    player_obj = matches[0]
                else:
                    # no Players were found
                    player_obj = None
    except (KeyError, IndexError, ValueError):
        pass
    if player_obj is None:
        raise CommandError("Invalid Player")
    elif not spectators and player_obj.world_object is None:
        raise CommandError("Invalid Spectator")
    return player_obj


def get_team(connection, value):
    value = value.lower()
    team_1 = connection.protocol.team_1
    team_2 = connection.protocol.team_2
    team_spectator = connection.protocol.team_spectator
    if value == team_1.name.lower():
        return connection.protocol.team_1
    elif value == team_2.name.lower():
        return connection.protocol.team_2
    elif value == team_spectator.name.lower():
        return connection.protocol.spectator_team
    elif value == '1':
        return connection.protocol.team_1
    elif value == '2':
        return connection.protocol.team_2
    elif value == 'spec':
        return connection.protocol.spectator_team
    raise ValueError('Invalid Team')


def join_arguments(arg, default=None):
    if not arg:
        return default
    return ' '.join(arg)


def parse_maps(pre_maps):
    maps = []
    for n in pre_maps:
        if n[0] == "#" and len(maps) > 0:
            maps[-1] += " " + n
        else:
            maps.append(n)

    return maps, ', '.join(maps)


def get_truthy(value):
    value = value.lower()

    if value in ("yes", "y", "on", 1):
        return True
    elif value in ("no", "n", "off", 0):
        return False
    else:
        return None


def handle_command(connection, command, parameters):
    """
    Public facing function to run a command, given the connection, a command
    name, and a list of parameters.

    Will log the command.
    """
    result = _handle_command(connection, command, parameters)

    if result == False:
        parameters = ['***'] * len(parameters)
    log_message = '<{}> /{} {}'.format(connection.name, command,
                                       ' '.join(parameters))
    if result:
        log_message += ' -> %s' % result
    log.info(escape_control_codes(log_message))

    return result

def _handle_command(connection, command, parameters):
    command = command.lower()
    try:
        command_name = _alias_map.get(command, command)
        command_func = _commands[command_name]
    except KeyError:
        return 'Unknown command'

    if not has_permission(command_func, connection):
        return "You can't use this command"

    argspec = inspect.getfullargspec(command_func)
    min_params = len(argspec.args) - 1 - len(argspec.defaults or ())
    max_params = len(argspec.args) - 1 if argspec.varargs is None else None
    len_params = len(parameters)

    if (len_params < min_params
            or max_params is not None and len_params > max_params):
        return format_command_error(
            command_func, 'Invalid number of arguments')

    msg = None

    try:
        return command_func(connection, *parameters)
    # TODO: remove all of this catching. Commands should deal with invalid
    # parameters themselves. Instead, we should catch ALL Exceptions and
    # make an attempt at displaying them nicely in format_command_error
    except KeyError:
        msg = None  # 'Invalid command'
        traceback.print_exc()
    except TypeError:
        print('Command', command, 'failed with args:', parameters)
        traceback.print_exc()
        msg = 'Command failed'
    except CommandError as e:
        msg = str(e)
    except PermissionDenied as e:
        msg = 'You can\'t do that: {}'.format(str(e))
    except ValueError as e:
        msg = str(e) if e.args else "Invalid parameters"

    return format_command_error(command_func, msg)


def handle_input(connection, user_input):
    # for IRC and console
    return handle_command(connection, *parse_command(user_input))
